"""Entities shared across domains"""

# pylint: disable=invalid-name,too-few-public-methods

from enum import Enum


class KeyType(Enum):
    """
    An enum that represents what a key is to be used for, i.e. signature,
    encryption, or both.
    """
    BOTH = 0
    ENCRYPTION = 1
    SIGNATURE = 2
    OTHER = -1


class PublicKey(object):
    """
    Public key object signifying a public key belonging to an Organization,
    Directory, or Service
    """

    def __init__(self, public_key_data):
        self.id = public_key_data['id']
        self.active = public_key_data['active']
        self.created = public_key_data['date_created']
        self.expires = public_key_data['date_expires']
        self.public_key = public_key_data['public_key']

        try:
            self.key_type = KeyType.BOTH if not \
                public_key_data.get('key_type') else \
                KeyType(public_key_data['key_type'])

        except ValueError:
            self.key_type = KeyType.OTHER



""" Service Policy objects """
from enum import Enum

from launchkey.exceptions import InvalidFenceType, InvalidPolicyAttributes


# pylint: disable=too-few-public-methods
class Fence(object):
    """ Base class of all Fences """

    def __init__(self, name):
        self.name = name


# pylint: disable=too-few-public-methods
class Policy(object):
    """
    Base class for new policy objects
    """
    def __init__(self, fences):
        if not fences:
            fences = []

        for fence in fences:
            if not isinstance(fence, Fence):
                raise InvalidFenceType("Invalid Fence object. Fence must be "
                                       "one of the following: "
                                       "[\"GeoCircleFence\", \"Territory"
                                       "Fence\"]")
        self.fences = fences


class Requirement(Enum):
    """ Requirement Enum """
    AMOUNT = "AMOUNT"
    TYPES = "TYPES"
    COND_GEO = "COND_GEO"
    OTHER = "OTHER"


class AuthorizationResponsePolicy(Policy):
    """
    Represents the AMI Authorization Policy

    :param requirement: Requirement Enum representing the policy requirement
    :param amount: Integer representing amount of factors to enforce
    :param fences: List of fence objects
    :param inherence_required: Boolean. Whether to require inherence factor
    :param knowledge_required: Boolean. Whether to require knowledge factor
    :param possession_required: Boolean. Whether to require possession factor
    """

    # pylint: disable=too-many-arguments
    def __init__(self, requirement=None, amount=0, fences=None,
                 inherence_required=False, knowledge_required=False,
                 possession_required=False):
        super().__init__(fences)

        if requirement and not isinstance(requirement, Requirement):
            raise InvalidPolicyAttributes("Requirement must be an enumeration "
                                          "of \"Requirement\"")

        if requirement and isinstance(requirement, Requirement):
            self.requirement = requirement

        else:
            self.requirement = Requirement.OTHER

        self.amount = amount
        self.inherence_required = inherence_required
        self.knowledge_required = knowledge_required
        self.possession_required = possession_required

    def to_dict(self):
        """
        returns the JSON representation of the auth response policy
        """
        return dict(self)

    def __repr__(self):
        return "AuthorizationResponsePolicy <" \
               "requirement={requirement}, " \
               "fences={fences}, " \
               "amount={amount}, " \
               "inherence_required={inherence_required}, " \
               "knowledge_required={knowledge_required}, " \
               "possession_required={possession_required}>". \
            format(
                requirement=repr(self.requirement),
                fences=repr(self.fences),
                amount=self.amount,
                inherence_required=self.inherence_required,
                knowledge_required=self.knowledge_required,
                possession_required=self.possession_required
            )

    def __iter__(self):
        yield "requirement", self.requirement.name
        yield "fences", [dict(fence) for fence in self.fences]
        yield "amount", self.amount
        yield "inherence_required", self.inherence_required
        yield "knowledge_required", self.knowledge_required
        yield "possession_required", self.possession_required


class ConditionalGeoFencePolicy(Policy):
    """
    Provides an Auth Request the ability to have the device based on user
    location implement differing Auth Policies

    :param inside: Auth Policy to execute if the device is inside
    the fence(s)
    :param outside: Auth Policy to execute if the device is outside
    :param deny_rooted_jailbroken: Deny request if device reports that it is
    rooted/jailbroken
    :param deny_emulator_simulator: Deny request if device reports it is a
    simulator or an emulator
    :param fences: list.  List of fences that the device should check
    the fence(s)
    """

    #  pylint: disable=too-many-arguments
    def __init__(self, inside, outside, deny_rooted_jailbroken=False,
                 deny_emulator_simulator=False, fences=None):
        super().__init__(fences)

        self.deny_rooted_jailbroken = deny_rooted_jailbroken
        self.deny_emulator_simulator = deny_emulator_simulator

        if inside.fences or outside.fences:
            raise InvalidPolicyAttributes("Fences are not allowed on Inside or"
                                          " Outside Policy objects")

        if inside.deny_rooted_jailbroken not in [False, None] or \
                outside.deny_rooted_jailbroken not in [False, None]:
            raise InvalidPolicyAttributes(
                "Setting deny_rooted_jailbroken is not allowed on Inside or "
                "Outside Policy objects"
            )

        if inside.deny_emulator_simulator not in [False, None] or \
                outside.deny_emulator_simulator not in [False, None]:
            raise InvalidPolicyAttributes(
                "Setting deny_emulator_simulator is not allowed on Inside or "
                "Outside Policy objects"
            )

        self.inside = self.__create_inner_policies(inside)
        self.outside = self.__create_inner_policies(outside)

    def to_dict(self):
        """
        returns the JSON representation of the policy object
        """
        return dict(self)

    def __repr__(self):
        return "ConditionalGeoFencePolicy <" \
               "inside={inside}, " \
               "outside={outside}, " \
               "deny_rooted_jailbroken={deny_rooted_jailbroken}, " \
               "deny_emulator_simulator={deny_emulator_simulator}, " \
               "fences={fences}>". \
            format(
                inside=repr(self.inside),
                outside=repr(self.outside),
                deny_rooted_jailbroken=self.deny_rooted_jailbroken,
                deny_emulator_simulator=self.deny_emulator_simulator,
                fences=repr(self.fences)
            )

    def __iter__(self):
        yield "type", "COND_GEO"
        yield "fences", [dict(fence) for fence in self.fences]
        yield "inside", dict(self.inside)
        yield "outside", dict(self.outside)
        yield "deny_rooted_jailbroken", self.deny_rooted_jailbroken
        yield "deny_emulator_simulator", self.deny_emulator_simulator

    # noinspection PyTypeChecker
    @staticmethod
    def __create_inner_policies(policy):
        if isinstance(policy, MethodAmountPolicy):
            new_policy = MethodAmountPolicy(
                amount=policy.amount, fences=None,
                deny_emulator_simulator=None, deny_rooted_jailbroken=None
            )
        elif isinstance(policy, FactorsPolicy):
            new_policy = FactorsPolicy(
                deny_rooted_jailbroken=None,
                deny_emulator_simulator=None,
                inherence_required=policy.inherence_required,
                knowledge_required=policy.knowledge_required,
                possession_required=policy.possession_required,
                fences=None
            )
        else:
            raise InvalidPolicyAttributes(
                "Inside and Outside policies must be one of the following: ["
                "\"FACTORS\", \"METHOD_AMOUNT\"]"
            )

        return new_policy


class LegacyPolicy(Policy):
    """
    Auth policy that loosely models the soon-to-be-deprecated
    `ServiceSecurityPolicy`

    :param amount: Integer amount of auth methods to require
    :param inherence_required: Boolean. Whether to require inherence factor
    :param knowledge_required: Boolean. Whether to require knowledge factor
    :param possession_required: Boolean. Whether to require possession factor
    :param deny_rooted_jailbroken: Boolean, deny request if device reports
    that it is rooted/jailbroken
    :param fences: List of GeoCircleFences
    :param time_restrictions: List of TimeFences
    """

    # pylint: disable=too-many-arguments
    def __init__(self, amount=0, inherence_required=False,
                 knowledge_required=False, possession_required=False,
                 deny_rooted_jailbroken=False, fences=None,
                 time_restrictions=None):

        sanitized_fences = []
        for fence in fences:
            if not isinstance(fence, GeoCircleFence):
                raise InvalidFenceType("A LegacyPolicy may only contain a "
                                       "GeoCircleFence.")

            sanitized_fences.append(fence)

        super().__init__(sanitized_fences)
        self.amount = amount
        self.inherence_required = inherence_required
        self.knowledge_required = knowledge_required
        self.possession_required = possession_required
        self.deny_rooted_jailbroken = deny_rooted_jailbroken
        self.fences = [] if not fences else fences
        self.time_restrictions = [] if not time_restrictions else \
            time_restrictions

    def to_dict(self):
        """
        returns the JSON representation of the legacy policy object
        """
        return dict(self)

    def __repr__(self):
        return "LegacyPolicy <" \
               "amount={amount}, " \
               "inherence_required={inherence_required}, " \
               "knowledge_required={knowledge_required}, " \
               "possession_required={possession_required}, " \
               "deny_rooted_jailbroken={deny_rooted_jailbroken}, " \
               "fences={fences}, " \
               "time_restrictions={time_restrictions}>". \
            format(
                amount=self.amount,
                inherence_required=self.inherence_required,
                knowledge_required=self.knowledge_required,
                possession_required=self.possession_required,
                deny_rooted_jailbroken=self.deny_rooted_jailbroken,
                fences=repr(self.fences),
                time_restrictions=repr(self.time_restrictions)
            )

    def __iter__(self):
        yield "type", "LEGACY"
        yield "amount", self.amount
        yield "inherence_required", self.inherence_required
        yield "knowledge_required", self.knowledge_required
        yield "possession_required", self.possession_required
        yield "deny_rooted_jailbroken", self.deny_rooted_jailbroken
        yield "fences", [dict(f) for f in self.fences]
        yield "time_restrictions", [dict(f) for f in self.time_restrictions]


class MethodAmountPolicy(Policy):
    """
    Auth policy object that handles authentication by the amount of factors

    :param amount: amount of methods to be used in this AuthRequest
    :param deny_rooted_jailbroken: Deny request if device reports that it is
    rooted/jailbroken
    :param deny_emulator_simulator: Deny request if device reports it is a
    simulator or an emulator
    :param fences: list.  List of fences that the device should check
    """
    def __init__(self, amount=0, deny_rooted_jailbroken=False,
                 deny_emulator_simulator=False, fences=None):
        super().__init__(fences)
        self.amount = amount
        self.deny_rooted_jailbroken = deny_rooted_jailbroken
        self.deny_emulator_simulator = deny_emulator_simulator

    def to_dict(self):
        """
        returns the JSON representation of the policy object
        """
        return dict(self)

    def __repr__(self):
        return "MethodAmountPolicy <" \
               "amount={amount}, " \
               "deny_rooted_jailbroken={deny_rooted_jailbroken}, " \
               "deny_emulator_simulator={deny_emulator_simulator}, " \
               "fences={fences}>". \
            format(
                amount=self.amount,
                deny_rooted_jailbroken=self.deny_rooted_jailbroken,
                deny_emulator_simulator=self.deny_emulator_simulator,
                fences=repr(self.fences)
            )

    def __iter__(self):
        yield "type", "METHOD_AMOUNT"
        yield "fences", [dict(fence) for fence in self.fences]
        yield "amount", self.amount
        if self.deny_rooted_jailbroken is not None:
            yield "deny_rooted_jailbroken", self.deny_rooted_jailbroken
        if self.deny_emulator_simulator is not None:
            yield "deny_emulator_simulator", self.deny_emulator_simulator


class FactorsPolicy(Policy):
    """
    Auth policy object that handles authentication based on type of factors

    Cannot be a mixed list of strings and Factor objects
    :param deny_rooted_jailbroken: Deny request if device reports that it is
    rooted/jailbroken
    :param deny_emulator_simulator: Deny request if device reports it is a
    simulator or an emulator
    :param inherence_required: Boolean. Whether to require inherence factor
    :param knowledge_required: Boolean. Whether to require knowledge factor
    :param possession_required: Boolean. Whether to require possession factor
    :param fences: List of fences to apply to the policy
    """

    # pylint: disable=too-many-arguments
    def __init__(self, deny_rooted_jailbroken=False,
                 deny_emulator_simulator=False, inherence_required=False,
                 knowledge_required=False, possession_required=False,
                 fences=None):
        super().__init__(fences)
        self.deny_rooted_jailbroken = deny_rooted_jailbroken
        self.deny_emulator_simulator = deny_emulator_simulator
        self.inherence_required = inherence_required
        self.knowledge_required = knowledge_required
        self.possession_required = possession_required

    def __factors_list(self):
        factors = []
        if self.inherence_required:
            factors.append("INHERENCE")

        if self.knowledge_required:
            factors.append("KNOWLEDGE")

        if self.possession_required:
            factors.append("POSSESSION")

        return factors

    def to_dict(self):
        """
        returns the JSON representation of the policy object
        """
        return dict(self)

    def __repr__(self):
        return "FactorsPolicy <" \
               "deny_rooted_jailbroken={deny_rooted_jailbroken}, " \
               "deny_emulator_simulator={deny_emulator_simulator}, " \
               "inherence_required={inherence_required}, " \
               "knowledge_required={knowledge_required}, " \
               "possession_required={possession_required}, " \
               "fences={fences}>". \
            format(
                deny_rooted_jailbroken=self.deny_rooted_jailbroken,
                deny_emulator_simulator=self.deny_emulator_simulator,
                inherence_required=self.inherence_required,
                knowledge_required=self.knowledge_required,
                possession_required=self.possession_required,
                fences=repr(self.fences)
            )

    def __iter__(self):
        yield "type", "FACTORS"
        yield "fences", [dict(fence) for fence in self.fences]
        yield "factors", self.__factors_list()
        if self.deny_rooted_jailbroken is not None:
            yield "deny_rooted_jailbroken", self.deny_rooted_jailbroken
        if self.deny_emulator_simulator is not None:
            yield "deny_emulator_simulator", self.deny_emulator_simulator


class GeoCircleFence(Fence):
    """
    Represents a fence based on latitude, longitude, and radius

    :param latitude: represents the distance north or south of the equator
    :param longitude: represents the distance east or west of the meridian
    :param radius: radius of circle from the point
    :param name: name of the Fence
    """
    def __init__(self, latitude, longitude, radius, name=None):
        super().__init__(name)
        self.latitude = float(latitude)
        self.longitude = float(longitude)
        self.radius = float(radius)

    def __repr__(self):
        return "GeoCircleFence <" \
               "latitude={latitude}, " \
               "longitude={longitude}, " \
               "radius={radius}, " \
               "name=\"{name}\">". \
            format(
                latitude=self.latitude,
                longitude=self.longitude,
                radius=self.radius,
                name=self.name
            )

    def __iter__(self):
        yield "type", "GEO_CIRCLE"
        yield "name", self.name
        yield "latitude", self.latitude
        yield "longitude", self.longitude
        yield "radius", self.radius

    def __eq__(self, other):
        if isinstance(other, GeoCircleFence):
            return self.latitude == other.latitude \
                and self.longitude == other.longitude \
                and self.radius == other.radius \
                and self.name == other.name

        return False


class TerritoryFence(Fence):
    """
    Represents a fence based on the territory

    :param country: two character representation of the country
    :param administrative_area:  ISO 3166-2 subdivision code ex: "US-CA"
    :param postal_code: string representation of area mail is delivered too
    :param name: name of the Fence
    """
    def __init__(self, country, administrative_area=None, postal_code=None,
                 name=None):
        super().__init__(name)
        self.country = country
        self.administrative_area = administrative_area
        self.postal_code = str(postal_code) if postal_code else None

    def __repr__(self):
        return "TerritoryFence <" \
               "country=\"{country}\", " \
               "administrative_area=\"{administrative_area}\", " \
               "postal_code=\"{postal_code}\", " \
               "name=\"{name}\">". \
            format(
                country=self.country,
                administrative_area=self.administrative_area,
                postal_code=self.postal_code,
                name=self.name
            )

    def __iter__(self):
        yield "type", "TERRITORY"
        yield "name", self.name
        yield "administrative_area", self.administrative_area
        yield "country", self.country
        yield "postal_code", self.postal_code

    def __eq__(self, other):
        if isinstance(other, TerritoryFence):
            return self.country == other.country \
                and self.administrative_area == other.administrative_area \
                and self.postal_code == other.postal_code \
                and self.name == other.name

        return False



"""Validators"""

from formencode import Schema, validators
from ..utils.validation import ValidateISODate


class AuthorizationInProgressValidator(Schema):
    """Pre-existing auth (Busy Signal) validation"""
    auth_request = validators.String()
    expires = ValidateISODate()
    from_same_service = validators.Bool()
    allow_extra_fields = True



"""Shared Validators"""


from formencode import FancyValidator, Invalid
from dateutil.parser import parse


class ValidateISODate(FancyValidator):
    """Validator for LaunchKey ISO dates"""

    @staticmethod
    def _to_python(value, state):
        try:
            val = parse(value)
        except ValueError:
            raise Invalid("Date/time format is invalid, it must be ISO 8601 "
                          "formatted  for UTZ with no offset (i.e. "
                          "2010-01-01T01:01:01Z)", value, state) from None
        return val



""" Organization Client """

# pylint: disable=too-many-public-methods,invalid-name,deprecated-method
# pylint: disable=too-many-arguments

from base64 import encodebytes
from .base import ServiceManagingBaseClient, api_call
from ..utils.shared import iso_format
from ..entities.shared import PublicKey
from ..entities.directory import Directory
from ..entities.validation import DirectoryValidator, PublicKeyValidator


class OrganizationClient(ServiceManagingBaseClient):
    """
    Organization Client for interacting with Organization endpoints
    """
    def __init__(self, subject_id, transport):
        super().__init__('org', subject_id, transport,
                         "/organization/v3/services")

    @api_call
    def create_directory(self, name):
        """
        Creates a new Directory
        :param name: Name describing the Directory that can be viewed in the
        Admin Center
        :return: String - ID of the Directory that is created
        :raise: launchkey.exceptions.DirectoryNameInUse - Directory name
        already taken
        """
        return self._transport.post("/organization/v3/directories",
                                    self._subject, name=name).data['id']

    @api_call
    def get_all_directories(self):
        """
        Retrieves all Directories belonging to an Organization
        :return: List - launchkey.entities.directory.Directory object
        containing Directory details
        """
        return [
            Directory(self._validate_response(directory, DirectoryValidator))
            for directory in
            self._transport.get("/organization/v3/directories",
                                self._subject).data]

    @api_call
    def get_directories(self, directory_ids):
        """
        Retrieves a list of Directories belonging to an Organization
        :param directory_ids: List of unique Directory IDs
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :return: List - launchkey.entities.directory.Directory object
        containing Directory details
        """
        return [
            Directory(self._validate_response(directory, DirectoryValidator))
            for directory in
            self._transport.post("/organization/v3/directories/list",
                                 self._subject,
                                 directory_ids=[str(directory_id) for
                                                directory_id in
                                                directory_ids]).data]

    @api_call
    def get_directory(self, directory_id):
        """
        Retrieves a Directory based on an input Directory ID
        :param directory_id: Unique Directory ID
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were not correct
        :return: launchkey.entities.directory.Directory object containing
        Directory details
        """
        return Directory(self._validate_response(
            self._transport.post("/organization/v3/directories/list",
                                 self._subject,
                                 directory_ids=[str(directory_id)]).data[0],
            DirectoryValidator))

    @api_call
    def update_directory(self, directory_id, ios_p12=False, android_key=False,
                         active=None, denial_context_inquiry_enabled=None,
                         webhook_url=False):
        """
        Updates a Directories's settings. If an optional parameter is not
        included it will not be updated.
        :param directory_id: Unique Directory ID
        :param ios_p12: MPNS P12 formatted push key containing both private
        key and cert (must be password free)
        :param android_key: GCM Push Key
        :param active: Boolean. Status preventing Directory Service Auths as
        well as other Directory related calls.
        :param denial_context_inquiry_enabled: Boolean. Should the user be
        prompted for denial context when they deny authorization requests
        for any and all child services.
        :param webhook_url: The URL to which web hooks for this Directory
        will be made.
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :return:
        """
        kwargs = {"directory_id": str(directory_id)}
        if ios_p12 is not False:
            kwargs['ios_p12'] = encodebytes(ios_p12).decode(
                'utf-8') if ios_p12 else ios_p12
        if android_key is not False:
            kwargs['android_key'] = android_key
        if active is not None:
            kwargs['active'] = active
        if denial_context_inquiry_enabled is not None:
            kwargs['denial_context_inquiry_enabled'] = \
                denial_context_inquiry_enabled
        if webhook_url is not False:
            kwargs['webhook_url'] = webhook_url
        self._transport.patch("/organization/v3/directories", self._subject,
                              **kwargs)

    @api_call
    def get_directory_public_keys(self, directory_id):
        """
        Retrieves a list of Public Keys belonging to a Directory
        :param directory_id: Unique Directory ID
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.Forbidden - The Directory you requested
        either does not exist or you do not have sufficient permissions.
        :return: List - launchkey.entities.shared.PublicKey
        """
        return [PublicKey(self._validate_response(key, PublicKeyValidator)) for
                key in
                self._transport.post("/organization/v3/directory/keys/list",
                                     self._subject,
                                     directory_id=str(directory_id)).data]

    # pylint: disable = duplicate-code
    @api_call
    def add_directory_public_key(self, directory_id, public_key, expires=None,
                                 active=None, key_type=None):
        """
        Adds a public key to an Directory
        :param directory_id: Unique Directory ID
        :param public_key: String RSA public key
        :param expires: Optional datetime.datetime stating a time in which
        the key will no longer be valid
        :param active: Optional bool stating whether the key should be
        considered active and usable.
        :param key_type: Optional KeyType enum to identify whether the key is
        an encryption key, signature key, or a dual use key
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were not correct
        :raise: launchkey.exceptions.InvalidPublicKey - The public key you
        supplied is not valid.
        :raise: launchkey.exceptions.PublicKeyAlreadyInUse - The public key
        you supplied already exists for the requested entity. It cannot be
        added again.
        :raise: launchkey.exceptions.Forbidden - The Directory you requested
        either does not exist or you do not have sufficient permissions.
        :return: MD5 fingerprint (key_id) of the public key,
        IE: e0:2f:a9:5a:76:92:6b:b5:4d:24:67:19:d1:8a:0a:75
        """
        kwargs = {"directory_id": str(directory_id), "public_key": public_key}
        if expires is not None:
            kwargs['date_expires'] = iso_format(expires)
        if active is not None:
            kwargs['active'] = active

        if key_type is not None:
            kwargs['key_type'] = key_type.value

        return \
            self._transport.post("/organization/v3/directory/keys",
                                 self._subject,
                                 **kwargs).data['key_id']

    @api_call
    def remove_directory_public_key(self, directory_id, key_id):
        """
        Removes a public key from a Directory. You may only remove a public
        key if other public keys exist. If you wish for a last remaining key
        to no longer be usable, use update_service_public_key to instead and
        set it as inactive.
        :param directory_id: Unique Directory ID
        :param key_id: MD5 fingerprint of the public key,
        IE: e0:2f:a9:5a:76:92:6b:b5:4d:24:67:19:d1:8a:0a:75
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were  not correct
        :raise: launchkey.exceptions.PublicKeyDoesNotExist - The key_id you
        supplied could not be found.
        :raise: launchkey.exceptions.LastRemainingKey - The last remaining
        public key cannot be removed.
        :raise: launchkey.exceptions.Forbidden - The Directory you requested
        either does not exist or you do not have sufficient permissions.
        :return:
        """
        self._transport.delete("/organization/v3/directory/keys",
                               self._subject, directory_id=str(directory_id),
                               key_id=key_id)

    @api_call
    def update_directory_public_key(self, directory_id, key_id, expires=False,
                                    active=None):
        """
        Removes a public key from a Directory
        :param directory_id: Unique Directory ID
        :param key_id: MD5 fingerprint of the public key,
        IE: e0:2f:a9:5a:76:92:6b:b5:4d:24:67:19:d1:8a:0a:75
        :param expires: datetime.datetime stating a time in which the key
        will  no longer be valid
        :param active: Bool stating whether the key should be considered
        active and usable
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were  not correct
        :raise: launchkey.exceptions.PublicKeyDoesNotExist - The key_id you
        supplied could not be found.
        :raise: launchkey.exceptions.Forbidden - The Directory you requested
        either does not exist or you do not have sufficient permissions.
        :return:
        """
        kwargs = {"directory_id": str(directory_id), "key_id": key_id}
        if active is not None:
            kwargs['active'] = active
        if expires is not False:
            kwargs['date_expires'] = iso_format(expires)
        self._transport.patch("/organization/v3/directory/keys", self._subject,
                              **kwargs)

    @api_call
    def generate_and_add_directory_sdk_key(self, directory_id):
        """
        Creates and retrieves a new Authenticator SDK Key for a Directory
        :param directory_id: Unique Directory ID
        :return: String - Newly generated Authenticator SDK Key
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were  not correct
        """
        return self._transport.post(
            "/organization/v3/directory/sdk-keys",
            self._subject, directory_id=str(directory_id)).data['sdk_key']

    @api_call
    def remove_directory_sdk_key(self, directory_id, sdk_key):
        """
        Removes an Authenticator SDK Key from a Directory
        :param directory_id: Unique Directory ID
        :param sdk_key: Authenticator SDK Key
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were not correct
        :raise: launchkey.exceptions.LastRemainingSDKKey - The last
        remaining SDK key cannot be removed
        :raise: launchkey.exceptions.InvalidSDKKey - The input SDK key does
        not belong to the given Directory
        :return:
        """
        self._transport.delete("/organization/v3/directory/sdk-keys",
                               self._subject, directory_id=str(directory_id),
                               sdk_key=sdk_key)

    @api_call
    def get_all_directory_sdk_keys(self, directory_id):
        """
        Retrieves a list of Authenticator SDK keys from a Directory
        :param directory_id: Unique Directory ID
        :raise: launchkey.exceptions.InvalidParameters - Input parameters
        were not correct
        :raise: launchkey.exceptions.EntityNotFound - Input Directory ID does
        not exist
        :return:
        """
        return self._transport.post(
            "/organization/v3/directory/sdk-keys/list",
            self._subject, directory_id=str(directory_id)
        ).data



"""Validators"""

# pylint: disable=too-few-public-methods

from formencode import Schema, validators, ForEach
from ..utils.validation import ValidateISODate


class PublicKeyValidator(Schema):
    """Public Key entity Validator"""
    id = validators.String()
    active = validators.Bool()
    date_created = ValidateISODate()
    date_expires = ValidateISODate()
    public_key = validators.String()
    key_type = validators.Int(if_missing=0, if_empty=0)
    allow_extra_fields = True


class DirectoryUserDeviceLinkResponseValidator(Schema):
    """Directory User Device link response validator"""
    qrcode = validators.String()  # URL
    code = validators.String(min=7)
    device_id = validators.String()
    allow_extra_fields = True


class DirectoryGetDeviceResponseValidator(Schema):
    """Directory get Device response validator"""
    id = validators.String()
    name = validators.String()
    status = validators.Int()
    type = validators.String()
    allow_extra_fields = True


class DirectoryGetSessionsValidator(Schema):
    """Directory get Sessions validator"""
    auth_request = validators.String()
    date_created = ValidateISODate()
    service_icon = validators.String()
    service_id = validators.String()
    service_name = validators.String()
    allow_extra_fields = True


class DirectoryValidator(Schema):
    """Directory entity validator"""
    id = validators.String()
    service_ids = ForEach(validators.String())
    sdk_keys = ForEach(validators.String())
    premium = validators.Bool()
    name = validators.String()
    android_key = validators.String()
    ios_certificate_fingerprint = validators.String()
    active = validators.Bool()
    denial_context_inquiry_enabled = validators.Bool(if_empty=False,
                                                     if_missing=False)
    webhook_url = validators.String()
    allow_extra_fields = True


class DirectoryDeviceLinkCompletionValidator(Schema):
    """Directory User Device link completion validator"""
    type = validators.OneOf(['DEVICE_LINK_COMPLETION'])
    device_id = validators.String()
    device_public_key = validators.String()
    device_public_key_id = validators.String()
    allow_extra_fields = True


class AuthorizationResponseValidator(Schema):
    """Authorization Response entity validator"""
    auth = validators.String()
    auth_jwe = validators.String(if_missing=None, if_empty=None)
    service_user_hash = validators.String()
    org_user_hash = validators.String()
    user_push_id = validators.String()
    public_key_id = validators.String()
    allow_extra_fields = True


class AuthorizationResponsePackageValidator(Schema):
    """Authorization Response Package entity validator"""
    service_pins = ForEach()
    auth_request = validators.String()  # UUID
    response = validators.Bool()
    device_id = validators.String()
    allow_extra_fields = True


class AuthMethodsValidator(Schema):
    """Auth methods validator"""
    method = validators.String()
    set = validators.Bool(if_empty=None)
    active = validators.Bool(if_empty=None)
    allowed = validators.Bool(if_empty=None)
    supported = validators.Bool(if_empty=None)
    user_required = validators.Bool(if_empty=None)
    passed = validators.Bool(if_empty=None)
    error = validators.Bool(if_empty=None)


class GeoFenceValidator(Schema):
    """ GeoFence Validator, can represent both GeoFence and GeoCircleFence """
    name = validators.String(if_missing=None)
    latitude = validators.Number()
    longitude = validators.Number()
    radius = validators.Number()


class GeoCircleFenceValidator(GeoFenceValidator):
    """ GeoFence Validator, can represent ONLY GeoCircleFence """
    type = validators.OneOf(["GEO_CIRCLE"])


class TerritoryFenceValidator(Schema):
    """ TerritoryFence Validator"""
    name = validators.String(if_missing=None)
    type = validators.OneOf(["TERRITORY"], if_missing=None)
    country = validators.Regex(r"^[A-Z]{2}$", not_empty=True)
    administrative_area = validators.Regex(r"^[A-Z]{2}-[A-Z]{2}[A-Z]?$",
                                           if_missing=None)
    postal_code = validators.String(if_missing=None, if_empty=None)

    @staticmethod
    def _validate_python(value, _state):
        if not value["administrative_area"]:
            del value["administrative_area"]

        if not value["postal_code"]:
            del value["postal_code"]


class FenceValidator(Schema):
    """Fence validator"""
    allow_extra_fields = True
    type = validators.OneOf(["GEO_CIRCLE", "TERRITORY"], if_missing=None)
    name = validators.String(if_missing=None)

    @staticmethod
    def _validate_python(value, _state):
        if not value["type"]:
            del value["type"]
            GeoFenceValidator().to_python(value)

        elif value["type"] == "GEO_CIRCLE":
            GeoCircleFenceValidator().to_python(value)

        elif value["type"] == "TERRITORY":
            TerritoryFenceValidator().to_python(value)


class AuthPolicyValidator(Schema):
    """Auth policy validate for auth method insights"""
    requirement = validators.String(if_missing=None, if_empty=None)
    amount = validators.Number(if_missing=None)
    types = ForEach(validators.String(), if_missing=None)
    geofences = ForEach(FenceValidator(), if_missing=[], if_empty=[])


class PolicyTerritoryValidator(Schema):
    """Validates Territory fences inside policies"""
    allow_extra_fields = True
    country = validators.String(not_empty=True)
    administrative_area = validators.String(if_missing=None)
    postal_code = validators.String(if_missing=None, if_empty=None)


class PolicyGeoCircleValidator(Schema):
    """Validates GeoCircle fences inside policies"""
    allow_extra_fields = True
    latitude = validators.Number(not_empty=True)
    longitude = validators.Number(not_empty=True)
    radius = validators.Number(not_empty=True)


class PolicyFenceValidator(Schema):
    """Validates fence objects in policies"""
    allow_extra_fields = True
    type = validators.String(not_empty=True)
    name = validators.String(if_missing=None, not_empty=True)

    @staticmethod
    def _validate_other(value, state):
        if "type" in value:
            if value["type"] == "TERRITORY":
                value.update(PolicyTerritoryValidator().to_python(
                    value, state))
            elif value["type"] == "GEO_CIRCLE":
                value.update(PolicyGeoCircleValidator().to_python(
                    value, state))
        return value


class ConditionalGeoFenceValidator(Schema):
    """Validates conditional geofence policies"""
    allow_extra_fields = True
    inside = validators.NotEmpty(accept_iterator=True)
    outside = validators.NotEmpty(accept_iterator=True)
    fences = ForEach(not_empty=True)

    @staticmethod
    def _validate_python(value, state):
        if 'inside' in value and 'outside' in value:
            value['inside'] = PolicyBaseValidator().to_python(
                value['inside'], state)
            value['outside'] = PolicyBaseValidator().to_python(
                value['outside'], state)
        return value


class MethodAmountPolicyValidator(Schema):
    """Validates method amount policies"""
    allow_extra_fields = True
    amount = validators.Int(not_empty=True)


class FactorsPolicyValidator(Schema):
    """Validates factors for policies"""
    allow_extra_fields = True
    factors = ForEach(validators.OneOf(
        ["KNOWLEDGE", "INHERENCE", "POSSESSION"]), not_empty=True)


class PolicyBaseValidator(Schema):
    """Base policy validator for legacy and new policies"""
    allow_extra_fields = True
    type = validators.String(if_missing="LEGACY")
    fences = ForEach(PolicyFenceValidator())

    @staticmethod
    def _validate_python(value, state):
        if value["type"] == "COND_GEO":
            value.update(ConditionalGeoFenceValidator().to_python(
                value, state))
        elif value["type"] == "METHOD_AMOUNT":
            value.update(MethodAmountPolicyValidator().to_python(value, state))
        elif value["type"] == "FACTORS":
            value.update(FactorsPolicyValidator().to_python(value, state))
        elif value["type"] == "LEGACY":
            if "deny_rooted_jailbroken" in value:
                del value["deny_rooted_jailbroken"]
            if "deny_emulator_simulator" in value:
                del value["deny_emulator_simulator"]
            del value["fences"]
        return value


class ServiceSecurityPolicyValidator(PolicyBaseValidator):
    """Service Policy validator"""
    allow_extra_fields = True
    deny_rooted_jailbroken = validators.Bool(if_missing=None)
    deny_emulator_simulator = validators.Bool(if_missing=None)


class JWEAuthorizationResponsePackageValidator(Schema):
    """Authorization Response JWE payload entity validator"""
    service_pins = ForEach()
    auth_request = validators.String()  # UUID
    type = validators.String()
    reason = validators.String()
    denial_reason = validators.String(if_missing=None, if_empty=None)
    device_id = validators.String()
    auth_policy = AuthPolicyValidator(if_missing=None)
    auth_methods = ForEach(AuthMethodsValidator())
    allow_extra_fields = True


class AuthorizeValidator(Schema):
    """Authorize entity validator"""
    auth_request = validators.String(not_empty=True)
    push_package = validators.String(if_missing=None, not_empty=True)
    device_ids = ForEach(validators.String(), if_missing=None)
    allow_extra_fields = True


class AuthorizeSSEValidator(Schema):
    """Authorize server-sent-event (webhook) validator"""
    service_user_hash = validators.String()
    api_time = validators.String()
    allow_extra_fields = True


class ServiceValidator(Schema):
    """Service entity validation"""
    id = validators.String()
    icon = validators.String()
    name = validators.String()
    description = validators.String()
    active = validators.Bool()
    callback_url = validators.String()
    allow_extra_fields = True


class ServiceTOTPVerificationValidator(Schema):
    """Service TOTP verification validation"""
    valid = validators.Bool()
    allow_extra_fields = True


class DirectoryUserTOTPValidator(Schema):
    """Directory TOTP post validator"""
    algorithm = validators.String()
    digits = validators.Int()
    period = validators.Int()
    secret = validators.String()
    allow_extra_fields = True



"""LaunchKey Service SDK module"""
SDK_VERSION = '4.0.1'
LAUNCHKEY_PRODUCTION = "https://api.launchkey.com"
VALID_JWT_ISSUER_LIST = ["svc", "dir", "org"]
JOSE_SUPPORTED_JWE_ALGS = ['RSA-OAEP']
JOSE_SUPPORTED_JWE_ENCS = ['A256CBC-HS512']
JOSE_SUPPORTED_JWT_ALGS = ["RS256", "RS384", "RS512"]
JOSE_SUPPORTED_CONTENT_HASH_ALGS = ["S256", "S384", "S512"]
JOSE_AUDIENCE = "lka"
JOSE_JWT_LEEWAY = 30
API_CACHE_TIME = 300



""" Transport for communicating with the LaunchKey API of HTTP"""

import requests

from .. import LAUNCHKEY_PRODUCTION
from .base import APIResponse, APIErrorResponse


class RequestsTransport(object):
    """
    Transport class for performing HTTP based queries using the requests
    library.
    """

    url = LAUNCHKEY_PRODUCTION
    testing = False
    verify_ssl = True
    allow_redirects = False

    def set_url(self, url, testing):
        """
        :param url: Base url for the querying LaunchKey API
        :param testing: Boolean stating whether testing mode is being
        performed. This will determine whether SSL should be verified.
        """
        self.url = url
        self.testing = testing
        self.verify_ssl = not self.testing

    @staticmethod
    def _parse_response(response):
        try:
            data = response.json()
        except ValueError:
            data = response.text

        try:
            response.raise_for_status()
            parsed_response = APIResponse(data, response.headers,
                                          response.status_code,
                                          raw_data=response.text)
        except requests.exceptions.HTTPError:
            if response.status_code < 500:
                parsed_response = APIErrorResponse(data, response.headers,
                                                   response.status_code,
                                                   response.reason,
                                                   response.text)
            else:
                raise

        return parsed_response

    def get(self, path, headers=None, data=None):
        """
        Performs an HTTP GET request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param headers: Headers to add onto the request
        :param data: Dictionary or bytes to be sent in the query string for
        the request.
        :return:
        """
        response = requests.get(self.url + path, params=data,
                                headers=headers, verify=self.verify_ssl,
                                allow_redirects=self.allow_redirects)
        return self._parse_response(response)

    def post(self, path, headers=None, data=None):
        """
        Performs and HTTP POST request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param headers: Headers to add onto the request
        :param data: Dictionary, bytes, or file-like object to send in the
        body of the request.
        :return:
        """
        response = requests.post(self.url + path, data=data,
                                 headers=headers, verify=self.verify_ssl,
                                 allow_redirects=self.allow_redirects)
        return self._parse_response(response)

    def put(self, path, headers=None, data=None):
        """
        Performs and HTTP PUT request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param headers: Headers to add onto the request
        :param data: Dictionary, bytes, or file-like object to send in the
        body of the request.
        :return:
        """
        response = requests.put(self.url + path, data=data,
                                headers=headers, verify=self.verify_ssl,
                                allow_redirects=self.allow_redirects)
        return self._parse_response(response)

    def delete(self, path, headers=None, data=None):
        """
        Performs and HTTP DELETE request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param headers: Headers to add onto the request
        :param data: Dictionary, bytes, or file-like object to send in the
        body of the request.
        :return:
        """
        response = requests.delete(self.url + path, data=data,
                                   headers=headers,
                                   verify=self.verify_ssl,
                                   allow_redirects=self.allow_redirects)
        return self._parse_response(response)

    def patch(self, path, headers=None, data=None):
        """
        Performs and HTTP PATCH request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param headers: Headers to add onto the request
        :param data: Dictionary, bytes, or file-like object to send in the
        body of the request.
        :return:
        """
        response = requests.patch(self.url + path, data=data,
                                  headers=headers, verify=self.verify_ssl,
                                  allow_redirects=self.allow_redirects)
        return self._parse_response(response)



"""Service Client"""

# pylint: disable=too-many-arguments

import warnings
from json import loads

from launchkey.exceptions import InvalidParameters, \
    UnableToDecryptWebhookRequest, UnexpectedAuthorizationResponse, \
    UnexpectedAPIResponse, UnexpectedWebhookRequest, XiovJWTValidationFailure,\
    XiovJWTDecryptionFailure
from launchkey.utils.shared import XiovJWTService, deprecated
from launchkey.entities.validation import AuthorizationResponseValidator, \
    AuthorizeSSEValidator, AuthorizeValidator, ServiceTOTPVerificationValidator
from launchkey.entities.service import AuthPolicy, AuthorizationResponse, \
    SessionEndRequest, AuthorizationRequest, AdvancedAuthorizationResponse, \
    DenialReason
from .base import BaseClient, api_call


class ServiceClient(BaseClient):
    """Service Client for interacting with Serive endpoints"""

    def __init__(self, subject_id, transport):
        super().__init__('svc', subject_id, transport)
        self.x_iov_jwt_service = XiovJWTService(self._transport, self._subject)

    @api_call
    def authorize(self, user, context=None, policy=None, title=None, ttl=None,
                  push_title=None, push_body=None):
        """
        Authorize a transaction for the provided user. This
        get_service_service method would be utilized if you are
        using this as a secondary factor for user login or authorizing a
        single transaction within your application.
        This will NOT begin a user session.
        :param user: LaunchKey Username, User Push ID, or Directory User ID
        for the End User
        :param context: Arbitrary string of data up to 400 characters to be
        presented to the End User during
        authorization to provide context regarding the individual request
        :param policy: Authorization policy override for this authorization.
        The policy can only increase the security level any existing policy
        in the Service Profile. It can never reduce the security level of
        the Service Profile's policy.
        :param title: String of data up to 200 characters to be presented to
        the End User during authorization as the title of the individual
        authorization request
        :param ttl: Time for this authorization request to be valid. If no
        value is provided, the system default will be used.
        :param push_title: Title that will appear in the mobile authenticator's
        push message. This feature is only available for Directory Services
        that have push credentials configured.
        :param push_body: Body that will appear in the mobile authenticator's
        push message. This feature is only available for Directory Services
        that have push credentials configured.
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.InvalidPolicyInput - Input policy was not
        valid
        :raise: launchkey.exceptions.PolicyFailure - Auth creation failed due
        to user not passing policy
        :raise: launchkey.exceptions.EntityNotFound - Username was invalid or
        the user does not have any valid devices
        :raise: launchkey.exceptions.RateLimited - Too many authorization
        requests have been created for this user
        :raise: launchkey.exceptions.InvalidPolicy - The input policy is not
        valid. It should be a
        launchkey.clients.service.AuthPolicy.
        Please wait and try again.
        :return: String - Unique identifier for tracking status of the
        authorization request
        """
        warnings.warn('This method has been deprecated and will be removed'
                      ' in a future major release!', DeprecationWarning)
        auth = self.authorization_request(user, context, policy, title, ttl,
                                          push_title, push_body)
        return auth.auth_request

    @api_call
    def authorization_request(self, user, context=None, policy=None,
                              title=None, ttl=None, push_title=None,
                              push_body=None, denial_reasons=None):
        """
        Authorize a transaction for the provided user. This get_service_service
        method would be utilized if you are using this as a secondary factor
        for user login or authorizing a single transaction within your
        application. This will NOT begin a user session.
        :param user: LaunchKey Username, User Push ID, or Directory User ID
        for the End User
        :param context: Arbitrary string of data up to 400 characters to be
        presented to the End User during authorization to provide context
        regarding the individual request
        :param policy: Authorization policy override for this authorization.
        The policy can only increase the security
        level any existing policy in the Service Profile. It can never reduce
        the security level of the Service Profile's policy.
        :param title: String of data up to 200 characters to be presented to
        the End User during authorization as the title of the individual
        authorization request
        :param ttl: Time for this authorization request to be valid. If no
        value is provided, the system default will be used.
        :param push_title: Title that will appear in the mobile authenticator's
        push message. This feature is only available for Directory Services
        that have push credentials configured.
        :param push_body: Body that will appear in the mobile authenticator's
        push message. This feature is only available for Directory Services
        that have push credentials configured.
        :param denial_reasons: List of denial reasons to present to the user if
        they deny the request. This list must include at least two items. At
        least one of the items must have a fraud value of false and at least
        one of the items must have a fraud value of true. If no denial_reasons
        are given the defaults will be used. If a list is provided and denial
        context inquiry is not enabled for the Directory, this request will
        error. This feature is only available for Directory Services.
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.InvalidPolicyInput - Input policy was not
        valid
        :raise: launchkey.exceptions.PolicyFailure - Auth creation failed due
        to user not passing policy
        :raise: launchkey.exceptions.EntityNotFound - Username was invalid or
        the user does not have any valid devices
        :raise: launchkey.exceptions.RateLimited - Too many authorization
        requests have been created for this user
        :raise: launchkey.exceptions.InvalidPolicy - The input policy is not
        valid. It should be a launchkey.clients.service.AuthPolicy.
        Please wait and try again.
        :raise: launchkey.exceptions.AuthorizationInProgress - Authorization
        request already exists for the requesting user. That request either
        needs to be responded to, expire out, or be canceled with
        cancel_authorization_request().
        :return AuthorizationResponse: Unique identifier for tracking status
        of the authorization request
        """
        kwargs = {'username': user}
        if context is not None:
            kwargs['context'] = context
        if title is not None:
            kwargs['title'] = title
        if ttl is not None:
            kwargs['ttl'] = ttl
        if push_title is not None:
            kwargs['push_title'] = push_title
        if push_body is not None:
            kwargs['push_body'] = push_body
        if policy is not None:
            if not isinstance(policy, AuthPolicy):
                raise InvalidParameters(
                    "Please verify the input policy is a "
                    "launchkey.entities.service.AuthPolicy class")
            kwargs['policy'] = policy.get_policy()
        if denial_reasons is not None:
            if not isinstance(denial_reasons, (list, set)):
                raise InvalidParameters(
                    "Please ensure that input denial_reasons are a list of "
                    "launchkey.entities.service.DenialReason classes.")
            parsed_reasons = []
            for reason in denial_reasons:
                if not isinstance(reason, DenialReason):
                    raise InvalidParameters(
                        "Please verify that denial_reasons are "
                        "launchkey.entities.service.DenialReason classes.")
                parsed_reasons.append(
                    {"id": reason.denial_id, "reason": reason.reason,
                     "fraud": reason.fraud}
                )
            kwargs['denial_reasons'] = parsed_reasons

        response = self._transport.post("/service/v3/auths",
                                        self._subject, **kwargs)
        data = self._validate_response(response, AuthorizeValidator)
        return AuthorizationRequest(data.get('auth_request'),
                                    data.get('push_package'),
                                    data.get('device_ids'))

    @api_call
    def get_advanced_authorization_response(self, authorization_request_id):
        """
        Request the response for a previous authorization call.
        :param authorization_request_id: Unique identifier returned by
        authorization_request()
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.RequestTimedOut - The authorization
        request has not been responded to before the
        timeout period (5 minutes)
        :raise: launchkey.exceptions.AuthorizationRequestCanceled - The
        authorization request has been canceled so a response cannot be
        retrieved.
        :return: None if the user has not responded otherwise a
        launchkey.entities.service.AdvancedAuthorizationResponse object
                 with the user's response
        in it
        """
        response = self._transport.get(
            "/service/v3/auths/%s" % authorization_request_id,
            self._subject)

        if response.status_code == 204:
            authorization_response = None
        else:
            data = self._validate_response(
                response,
                AuthorizationResponseValidator)
            authorization_response = AdvancedAuthorizationResponse(
                data,
                self._transport
            )

        return authorization_response

    @deprecated
    def get_authorization_response(self, authorization_request_id):
        """
        NOTE: This method is being deprecated. Use
        `get_advanced_authorization_response` instead!

        Request the response for a previous authorization call.
        :param authorization_request_id: Unique identifier returned by
        authorization_request()
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.RequestTimedOut - The authorization
        request has not been responded to before the
        timeout period (5 minutes)
        :raise: launchkey.exceptions.AuthorizationRequestCanceled - The
        authorization request has been canceled so a response cannot be
        retrieved.
        :return: None if the user has not responded otherwise a
        launchkey.entities.service.AuthorizationResponse object
                 with the user's response
        in it
        """
        advanced_authorization_response = \
            self.get_advanced_authorization_response(authorization_request_id)

        if not advanced_authorization_response:
            return None

        return AuthorizationResponse(
            advanced_authorization_response.data,
            advanced_authorization_response.transport)

    @api_call
    def cancel_authorization_request(self, authorization_request_id):
        """
        Request to cancel an authorization request for the End User
        :param authorization_request_id: Unique identifier returned by
        authorization_request()
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.EntityNotFound - The authorization
        request does not exist.
        :raise: launchkey.exceptions.AuthorizationRequestCanceled - The
        authorization request has already been canceled.
        :raise: launchkey.exceptions.AuthorizationResponseExists - The
        authorization request has already been responded to so it cannot be
        canceled.
        """
        self._transport.delete(
            "/service/v3/auths/%s" % authorization_request_id,
            self._subject)

    @api_call
    def session_start(self, user, authorization_request_id):
        """
        Request to start a Service Session for the End User which was derived
        from a authorization request
        :param user: LaunchKey Username, User Push ID, or Directory User ID for
        the End User
        :param authorization_request_id: Unique identifier returned by
        authorization_request()
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.EntityNotFound - The input username was
        not valid
        """
        self._transport.post("/service/v3/sessions",
                             self._subject,
                             username=user,
                             auth_request=authorization_request_id)

    @api_call
    def session_end(self, user):
        """
        Request to end a Service Session for the End User
        :param user: LaunchKey Username, User Push ID, or Directory User ID for
        the End User
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.EntityNotFound - The input username was
        not valid
        """
        self._transport.delete("/service/v3/sessions",
                               self._subject,
                               username=user)

    @api_call
    def verify_totp(self, user, otp):
        """
        Verifies a given TOTP is valid for a given user.
        :param user: Unique value identifying the End User in your
        system. This value was used to create the Directory User and Link
        Device.
        :param otp: 6-8 digit OTP code for to verify.
        :return: Boolean stating whether the given OTP code is valid.
        :raise: launchkey.exceptions.EntityNotFound - Unable to find TOTP
        configuration for given user.
        """
        response = self._transport.post("/service/v3/totp",
                                        self._subject, identifier=user,
                                        otp=otp)
        data = self._validate_response(
            response,
            ServiceTOTPVerificationValidator)

        return data["valid"]

    def handle_advanced_webhook(self, body, headers, method=None, path=None):
        """
        Handle an advanced webhook callback
        In the event of a Logout webhook, be sure to call session_end() when
        you complete the process of ending the user's session in your
        implementation.  This will remove the corresponding Application from
        the authorization list on all of the the user's mobile devices.
        :param body: The raw body that was send in the POST content
        :param headers: A generic map of response headers. These will be used
        to access and validate authorization
        :param path:  The path of the request
        :param method: The HTTP method of the request
        :return: launchkey.entities.service.SessionEndRequest or
        launchkey.entities.service.AdvancedAuthorizationResponse
        :raises launchkey.exceptions.UnexpectedWebhookRequest: when the
        request or its cannot be parsed or fails
        validation.
        :raises launchkey.exceptions.UnableToDecryptWebhookRequest: when the
        request is an authorization response webhook and the request body
        cannot be decrypted
        :raises launchkey.exceptions.UnexpectedAuthorizationResponse: when the
        decrypted auth package is missing required data. This error is
        indicative of a non webhook request being sent to the method.
        :raises launchkey.exceptions.UnexpectedKeyID: when the auth package in
        an authorization response webhook request body is decrypted using a
        public key whose private key is not known by the client. This can be
        a configuration issue.
        :raises launchkey.exceptions.UnexpectedDeviceResponse: when the auth
        package received from the device is invalid. This error is
        indicative of a man in the middle (MITM) attack.
        :raises launchkey.exceptions.WebhookAuthorizationError: when the
        "Authorization" header in the headers.
        """
        if method is None:
            warnings.warn("Not passing a valid request method string is "
                          "deprecated and will be required in the next "
                          "major version", PendingDeprecationWarning)

        if path is None:
            warnings.warn("Not passing a valid request path string is "
                          "deprecated and will be required in the next "
                          "major version", PendingDeprecationWarning)
        try:
            if "service_user_hash" in "%s" % body:
                body = self.x_iov_jwt_service.verify_jwt_request(body, headers,
                                                                 method, path)
                try:
                    body = self._validate_response(
                        loads(body),
                        AuthorizeSSEValidator)
                except UnexpectedAPIResponse as reason:
                    raise UnexpectedWebhookRequest(reason=reason) from reason
                result = SessionEndRequest(
                    body['service_user_hash'],
                    self._transport.parse_api_time(body['api_time']))
            else:
                try:
                    decrypted_body = self.x_iov_jwt_service.decrypt_jwe(
                        body, headers, method, path
                    )
                    auth_response = loads(decrypted_body)
                    result = AdvancedAuthorizationResponse(
                        auth_response,
                        self._transport
                    )
                except XiovJWTDecryptionFailure as reason:
                    raise UnableToDecryptWebhookRequest(reason=reason) \
                        from reason
                except KeyError as reason:
                    raise UnexpectedAuthorizationResponse(reason=reason) \
                        from reason
        except XiovJWTValidationFailure as reason:
            raise UnexpectedWebhookRequest(reason) from reason

        return result

    @deprecated
    def handle_webhook(self, body, headers, method=None, path=None):
        """
        NOTE: This method is being deprecated. Use `handle_advanced_webhook`
        instead!

        Handle a webhook callback
        In the event of a Logout webhook, be sure to call session_end() when
        you complete the process of ending the user's session in your
        implementation.  This will remove the corresponding Application from
        the authorization list on all of the the user's mobile devices.
        :param body: The raw body that was send in the POST content
        :param headers: A generic map of response headers. These will be used
        to access and validate authorization
        :param path:  The path of the request
        :param method: The HTTP method of the request
        :return: launchkey.entities.service.SessionEndRequest or
        launchkey.entities.service.AuthorizationResponse
        :raises launchkey.exceptions.UnexpectedWebhookRequest: when the
        request or its cannot be parsed or fails
        validation.
        :raises launchkey.exceptions.UnableToDecryptWebhookRequest: when the
        request is an authorization response webhook and the request body
        cannot be decrypted
        :raises launchkey.exceptions.UnexpectedAuthorizationResponse: when the
        decrypted auth package is missing required data. This error is
        indicative of a non webhook request being sent to the method.
        :raises launchkey.exceptions.UnexpectedKeyID: when the auth package in
        an authorization response webhook request body is decrypted using a
        public key whose private key is not known by the client. This can be
        a configuration issue.
        :raises launchkey.exceptions.UnexpectedDeviceResponse: when the auth
        package received from the device is invalid. This error is
        indicative of a man in the middle (MITM) attack.
        :raises launchkey.exceptions.WebhookAuthorizationError: when the
        "Authorization" header in the headers.
        """
        advanced_authorization_response = self.handle_advanced_webhook(
            body, headers, method, path)

        if isinstance(advanced_authorization_response, SessionEndRequest):
            return advanced_authorization_response

        return AuthorizationResponse(
            advanced_authorization_response.data,
            advanced_authorization_response.transport)



""" Module for shared code amongst transports """

# pylint: disable=too-few-public-methods, too-many-arguments


class APIResponse(object):
    """ Response from API """
    data = None
    headers = None
    status_code = None
    reason = None

    def __init__(self, data, headers, status_code, reason=None, raw_data=None):
        self.data = data
        self.headers = headers
        self.status_code = status_code
        self.reason = reason
        self.raw_data = raw_data

    def __str__(self):
        return super().__str__() + ": %s %s %s" % (
            self.status_code, self.reason, self.data)


class APIErrorResponse(APIResponse):
    """Response that should be raised when an OK status is not found"""



"""Directory level entities"""

# pylint: disable=invalid-name,too-few-public-methods,
# pylint: disable=too-many-instance-attributes


from launchkey.exceptions import InvalidDeviceStatus


class Directory(object):
    """Directory entity"""
    def __init__(self, data):
        self.service_ids = data['service_ids']
        self.sdk_keys = data['sdk_keys']
        self.premium = data['premium']
        self.name = data['name']
        self.android_key = data['android_key']
        self.ios_certificate_fingerprint = data['ios_certificate_fingerprint']
        self.active = data['active']
        self.id = data['id']
        self.denial_context_inquiry_enabled = \
            data['denial_context_inquiry_enabled']
        self.webhook_url = data['webhook_url']

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, Directory):
            eq = self.service_ids == other.service_ids and \
                      self.sdk_keys == other.sdk_keys and \
                      self.premium == other.premium and \
                      self.name == other.name and \
                      self.android_key == other.android_key and \
                      self.ios_certificate_fingerprint == \
                      other.ios_certificate_fingerprint and \
                      self.active == other.active and \
                      self.id == other.id and \
                      self.denial_context_inquiry_enabled == \
                      other.denial_context_inquiry_enabled and \
                      self.webhook_url == other.webhook_url
        else:
            eq = False
        return eq

    def __repr__(self):
        return "Directory <id=\"{id}\", name=\"{name}\", " \
               "service_ids={service_ids}, sdk_keys={sdk_keys}, " \
               "premium={premium}, " \
               "ios_certificate_fingerprint=" \
               "\"{ios_certificate_fingerprint}\", " \
               "active={active}, " \
               "denial_context_inquiry_enabled=" \
               "{denial_context_inquiry}, webhook_url=\"{webhook_url}\">". \
            format(
                id=self.id,
                name=self.name,
                service_ids=self.service_ids,
                sdk_keys=self.sdk_keys,
                premium=self.premium,
                ios_certificate_fingerprint=self.ios_certificate_fingerprint,
                active=self.active,
                denial_context_inquiry=self.denial_context_inquiry_enabled,
                webhook_url=self.webhook_url
            )


class Session(object):
    """User Service Session"""

    def __init__(self, session):
        self.auth_request = session['auth_request']
        self.created = session['date_created']
        self.service_icon = session['service_icon']
        self.service_id = session['service_id']
        self.service_name = session['service_name']


class DirectoryUserDeviceLinkData(object):
    """Directory user device data used to finish the linking process"""

    def __init__(self, data):
        self.qrcode = data['qrcode']
        self.code = data['code']
        self.device_id = data['device_id']


class DirectoryUserTOTP(object):
    """Directory user TOTP data that is returned after an add request"""

    def __init__(self, data):
        self.secret = data['secret']
        self.algorithm = data['algorithm']
        self.period = data['period']
        self.digits = data['digits']

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, DirectoryUserTOTP):
            eq = self.secret == other.secret and \
                 self.algorithm == other.algorithm and \
                 self.period == other.period and \
                 self.digits == other.digits
        else:
            eq = False
        return eq

    def __repr__(self):
        return "DirectoryUserTOTP <secret=\"{secret}\", " \
               "algorithm=\"{algorithm}\", period={period}, " \
               "digits={digits}>". \
            format(
                secret=self.secret,
                algorithm=self.algorithm,
                period=self.period,
                digits=self.digits
            )


class DeviceLinkCompletionResponse(object):
    """Package returned on the event that a device finishes linking"""

    def __init__(self, data):
        self.device_id = data['device_id']
        self.device_public_key = data['device_public_key']
        self.device_public_key_id = data['device_public_key_id']


class DeviceStatus(object):
    """Activation status of a directory user's device"""

    _status_map = {
        0: ("LINK_PENDING", False),
        1: ("LINKED", True),
        2: ("UNLINK_PENDING", True)
    }

    def __init__(self, status_code):
        if status_code not in self._status_map:
            raise InvalidDeviceStatus(
                "Status code %s was not recognized" % status_code)
        self._status_code = status_code

    @property
    def is_active(self):
        """
        Is the device active?
        :return: active status
        """
        return self._status_map[self._status_code][1]

    @property
    def status_code(self):
        """
        What's the status iof the device
        :return: Status code
        """
        return self._status_map[self._status_code][0]

    def __repr__(self):
        return \
            "DeviceStatus <status_code=\"{status_code}\", " \
            "is_active={is_active}>".format(
                status_code=self.status_code,
                is_active=self.is_active
            )


class Device(object):
    """Device object belonging to a directory user"""

    def __init__(self, data):
        self.id = data['id']
        self.name = data['name']
        self.status = DeviceStatus(data['status'])
        self.type = data['type']

    def __repr__(self):
        return "Device <id=\"{id}\", name=\"{name}\", " \
               "status={status}, type=\"{type}\">". \
            format(
                id=self.id,
                name=self.name,
                status=self.status,
                type=self.type
            )



"""Organization Factory"""

# pylint: disable=too-many-arguments

from .. import LAUNCHKEY_PRODUCTION
from ..clients import DirectoryClient, OrganizationClient, ServiceClient
from .base import BaseFactory


class OrganizationFactory(BaseFactory):
    """
    Factory for creating clients when representing a LaunchKey Organization
    """

    def __init__(self, organization_id, private_key, url=LAUNCHKEY_PRODUCTION,
                 testing=False, transport=None):
        """
        :param organization_id: UUID for the requesting organization
        :param private_key: PEM formatted private key string
        :param url: URL for the LaunchKey API
        :param testing: Boolean stating whether testing mode is being used.
        This will determine whether SSL validation occurs.
        :param: transport: Instantiated transport object. The default and
        currently only supported transport is
        launchkey.transports.JOSETransport. If you wish to set encryption or
        hashing algorithms, this is where you would do it. IE:
        JOSETransport(jwt_algorithm="RS512", jwe_cek_encryption="RSA-OAEP",
        jwe_claims_encryption="A256CBC-HS512", content_hash_algorithm="S256")
        """
        super().__init__('org', organization_id,
                         private_key, url, testing,
                         transport)

    def make_directory_client(self, directory_id):
        """
        Retrieves a client to make directory calls.
        :param directory_id: Directory id
        :return: launchkey.clients.DirectoryClient
        """
        return DirectoryClient(directory_id, self._transport)

    def make_organization_client(self):
        """
        Retrieves a client to make organization calls.
        :return: launchkey.clients.OrganizationClient
        """
        return OrganizationClient(self._issuer_id, self._transport)

    def make_service_client(self, service_id):
        """
        Retrieves a client to make service calls.
        :param service_id: Service id
        :return: launchkey.clients.ServiceClient
        """
        return ServiceClient(service_id, self._transport)



""" JOSE based transport"""

# pylint: disable=too-many-instance-attributes, too-many-arguments

import warnings
import json
import platform

from base64 import b64decode
from uuid import UUID, uuid4
from hashlib import sha256, sha384, sha512, md5
from time import time
from calendar import timegm
from dateutil.parser import parse
from Cryptodome.PublicKey import RSA
from Cryptodome.Cipher import PKCS1_OAEP
from jwkest import JWKESTException
from jwkest.jwk import RSAKey, import_rsa_key
from jwkest.jws import JWS, NoSuitableSigningKeys
from jwkest.jwe import JWE, JWEnc
from jwkest.jwt import JWT, BadSyntax

from ..exceptions import InvalidEntityID, InvalidPrivateKey, \
    InvalidIssuer, InvalidAlgorithm, LaunchKeyAPIException, \
    JWTValidationFailure, UnexpectedAPIResponse, NoIssuerKey, \
    InvalidJWTResponse, UnexpectedKeyID, EntityKeyNotFound
from .. import VALID_JWT_ISSUER_LIST, API_CACHE_TIME, \
    JOSE_SUPPORTED_CONTENT_HASH_ALGS, JOSE_SUPPORTED_JWE_ALGS, \
    JOSE_SUPPORTED_JWE_ENCS, JOSE_SUPPORTED_JWT_ALGS, \
    JOSE_AUDIENCE, JOSE_JWT_LEEWAY, SDK_VERSION
from .http import RequestsTransport
from .base import APIErrorResponse


class JOSETransport(object):
    """
    Transport to wrap HTTP transport for providing request/response validation,
    encryption, and decryption.
    """
    audience = JOSE_AUDIENCE

    def __init__(self, jwt_algorithm="RS512", jwe_cek_encryption="RSA-OAEP",
                 jwe_claims_encryption="A256CBC-HS512",
                 content_hash_algorithm="S256", http_client=None):
        """
        :param jwt_algorithm: JWT Signing algorithm
                              Currently supported: RS256, RS384, RS512
        :param jwe_cek_encryption: JWE algorithm to use with CEK encryption
        Currently supported: RSA-OAEP
        To be supported: RSA-OAEP-256 (pending pyjwkest release to > 1.3.2)
        :param jwe_claims_encryption: JWE algorithm to use for claims
        encryption.
        Currently supported: A256CBC-HS512
        :param content_hash_algorithm: Hashing algorithm for signing
        content body.
        Currently supported: S256, S384, S512 (shortened forms of SHAxxx)
        :param http_client: HTTP transport to contact the LaunchKey api after
        JOSE processing is complete
        """
        self.issuer = None
        self.issuer_id = None
        self.signing_key = None
        self.loaded_issuer_private_keys = {}
        self.issuer_private_keys = []
        self._server_time_difference = None, None

        # Single key ID with timestamp of when it was set.
        self._current_kid = None, None

        # Dictionary of public keys with `kid` being the dict key
        # and the public key being the value
        self._public_key_cache = {}

        self.jwt_algorithm = self.__verify_supported_algorithm(
            jwt_algorithm, JOSE_SUPPORTED_JWT_ALGS)
        self.jwe_cek_encryption = self.__verify_supported_algorithm(
            jwe_cek_encryption, JOSE_SUPPORTED_JWE_ALGS)
        self.jwe_claims_encryption = self.__verify_supported_algorithm(
            jwe_claims_encryption, JOSE_SUPPORTED_JWE_ENCS)
        self.content_hash_algorithm = self.__verify_supported_algorithm(
            content_hash_algorithm,
            JOSE_SUPPORTED_CONTENT_HASH_ALGS)
        self._http_client = http_client \
            if http_client is not None else RequestsTransport()

    @staticmethod
    def __verify_supported_algorithm(algorithm, supported_list):
        """
        Verifies an input string algorithm is in an input list, and raises
        the appropriate exception if it is not
        """
        if algorithm not in supported_list:
            raise InvalidAlgorithm(
                "Input algorithm {0} is not in "
                "the supported list of {1}".format(algorithm, supported_list))
        return algorithm

    @staticmethod
    def __get_jti():
        """Retrieves a unique JWT ID"""
        return str(uuid4())

    @staticmethod
    def __generate_key_id(rsa_key):
        """
        Generates a key id to be used in JWE + JWT. It is based on the digest
        of the key.
        :param rsa_key: RSA private key in string format
        :return:
        """
        key = RSA.importKey(rsa_key)
        md5digest = md5(key.publickey().exportKey('DER')).hexdigest()
        return ":".join(
            md5digest[i:i + 2] for i in range(0, len(md5digest), 2))

    @staticmethod
    def __get_kid_from_api_response_headers(headers):
        """
        Gets `kid` property from a JWT token within the headers of a
        LaunchKey API response.
        :param headers: Response headers
        :return: string of the `kid`
        :raises launchkey.exceptions.UnexpectedAPIResponse: if unable to unpack
            the JWT or if the JWT header does not exist
        :raises launchkey.exceptions.JWTValidationFailure: if `kid` is missing
            or invalid
        """
        try:
            jwt = headers.get("X-IOV-JWT")
            jwt_headers = JWT().unpack(jwt).headers
        except (BadSyntax, IndexError, ValueError) as jwt_error:
            raise UnexpectedAPIResponse("JWT was missing or malformed in API "
                                        "response.") from jwt_error

        kid = jwt_headers.get("kid")

        if not isinstance(kid, str):
            raise JWTValidationFailure("`kid` header in JWT was missing or"
                                       " invalid.")

        return kid

    @staticmethod
    def _get_kid_from_api_response(response):
        """
        Gets `kid` property from a JWT token within a LaunchKey API
        response.
        :param response: Response object
        :return: string of the `kid`
        """
        try:
            return response.headers["X-IOV-KEY-ID"]
        except KeyError:
            raise UnexpectedAPIResponse("X-IOV-KEY-ID was missing or malformed"
                                        " in API response.") from None

    @staticmethod
    def parse_api_time(api_time):
        """
        Parses LaunchKey api_time to a unix timestamp
        :return: int representation of the timestamp
        """
        return timegm(parse(api_time).timetuple())

    @property
    def server_time_difference(self):
        """
        The time drag between the sdk and the Launchkey API. The result is
        cached for API_CACHE_TIME
        """
        now = int(time())
        if self._server_time_difference[1] is None or now - \
                self._server_time_difference[1] > API_CACHE_TIME:
            response = self.get("/public/v3/ping", None)
            try:
                self._server_time_difference = now - self.parse_api_time(
                    response.data['api_time']), now
            except (KeyError, ValueError, TypeError):
                raise UnexpectedAPIResponse(
                    "Unexpected api time received: %s" % response.data) \
                    from None
        return self._server_time_difference[0]

    @property
    def api_public_keys(self):
        """
        List of RSA keys that have been generated from public keys supplied
        by the LaunchKey API. If no keys exist in the public key cache,
        then fetch from LaunchKey API
        :return: List of RSAKeys
        """
        if not self._public_key_cache:
            self.update_and_return_active_encryption_kid()
        return list(self._public_key_cache.values())

    def update_and_return_active_encryption_kid(self):
        """
        Determines whether a new current key is necessary, and if so, retrieves
        new `kid` and public key from LaunchKey API, sets the current `kid`
        with current timestamp, and caches the new key.
        :return: Currently active key id
        """
        now = int(time())
        current_kid, current_kid_timestamp = self._current_kid

        if (not current_kid or not isinstance(current_kid_timestamp, int)) \
                or now - current_kid_timestamp > API_CACHE_TIME:
            new_kid, new_public_key = self._get_current_kid_and_key()
            self._current_kid = new_kid, now
            self._cache_public_key(new_kid, new_public_key)
        return self._current_kid[0]

    def _get_key_by_kid(self, kid):
        """
        Gets public key from LaunchKey API by `kid` string.
        :param kid: string of the `kid`
        :return: string of the public key
        """
        response = self.get("/public/v3/public-key/%s" % kid)
        return self._handle_public_key_api_response(response)[1]

    def _get_current_kid_and_key(self):
        """
        Gets the current `kid` public key from LaunchKey API.
        :return: tuple with string of the `kid` and string of the public key
        """
        response = self.get("/public/v3/public-key")
        return self._handle_public_key_api_response(response)

    def _handle_public_key_api_response(self, response):
        """
        Parses a LaunchKey public key API response.
        :return: tuple with string of the `kid` and string of the public key
        :raises UnexpectedAPIResponse: if the LaunchKey API returns a 404
            or if the response object contains no data
        """
        if response.status_code == 404:
            raise UnexpectedAPIResponse("Key was not found.")

        kid = self._get_kid_from_api_response(response)
        public_key = response.data

        if not isinstance(public_key, str):
            raise UnexpectedAPIResponse("Unexpected API public key response"
                                        " received: %s" % response.data)

        return kid, public_key

    def _cache_public_key(self, kid, public_key):
        """
        Generate an RSAKey with the public key, and store within the public
        key cache. This only occurs if an RSAKey has not already been cached
        for the given `kid` and public key.
        :param kid: string of the `kid`
        :param public_key: string of the public key
        :return:
        """
        if not self._public_key_cache.get(kid):
            try:
                rsa_key = RSAKey(key=import_rsa_key(public_key), kid=kid)
            except (TypeError, ValueError) as rsa_error:
                raise UnexpectedAPIResponse("RSA parsing error for public key"
                                            ": %s" % public_key) from rsa_error

            self._public_key_cache[kid] = rsa_key

    def _find_key_by_kid(self, kid):
        """
        Finds a public key within the public key cache given a `kid`. If
        none exists, retrieves from the LaunchKey API by `kid`.
        :param kid: string of the `kid`
        :return: string of the public key
        """
        key = self._public_key_cache.get(kid)
        if not key:
            return self._get_key_by_kid(kid)

        return key

    def add_encryption_private_key(self, private_key):
        """
        Adds a private key to the list of keys available for decryption
        :return: Boolean - Whether the key is already in the list
        """
        new_key = RSAKey(key=import_rsa_key(private_key),
                         kid=self.__generate_key_id(private_key))
        for key in self.issuer_private_keys:
            if new_key.kid == key.kid:
                return False
        self.issuer_private_keys.append(new_key)
        self.loaded_issuer_private_keys[new_key.kid] = PKCS1_OAEP.new(
            RSA.importKey(private_key))
        return True

    def add_issuer_key(self, private_key):
        """
        Adds a private key to the list of keys available for decryption
        :return: Boolean - Whether the key is already in the list
        """
        warnings.warn(
            "This method will be removed in the future. Please use "
            "add_encryption_key instead.",
            DeprecationWarning)
        return self.add_encryption_private_key(private_key)

    def set_url(self, url, testing):
        """
        Creates a new http_client using the input url and testing flag
        :param url: Base url for the querying LaunchKey API
        :param testing: Boolean stating whether testing mode is being
        performed. This will determine whether SSL should be verified.
        :return:
        """
        self._http_client.set_url(url, testing)

    def set_issuer(self, issuer, issuer_id, private_key):
        """
        Set the issuer credentials
        :param issuer: Issuer entity type (svc, dir, or org)
        :param issuer_id: Identifier for the issuer entity
        :param private_key: PEM formatted private key for issuer entity which
                            will be used for signing requests.
        :return: None
        :raises launchkey.exceptions.InvalidEntityID: when issuer_id is not
        valid
        :raises launchkey.exceptions.InvalidIssuer: when issuer is not valid
        :raises launchkey.exceptions.InvalidPrivateKey when private_key is
        not valid
        """
        if issuer not in VALID_JWT_ISSUER_LIST:
            raise InvalidIssuer("Invalid issuer it must be one of: {0}".format(
                VALID_JWT_ISSUER_LIST))
        try:
            UUID(str(issuer_id))
        except ValueError:
            raise InvalidEntityID(
                "The given id was invalid. Please ensure it is a UUID.") \
                from None
        self.issuer = "%s:%s" % (issuer, issuer_id)
        try:
            self.signing_key = RSAKey(
                key=import_rsa_key(private_key),
                kid=self.__generate_key_id(private_key)
            )
        except ValueError as rsa_error:
            raise InvalidPrivateKey(
                "Invalid private key. Please ensure you are submitting "
                "a string representation of a PEM private key.") from rsa_error

    def _get_jwt_signature(self, params):
        try:
            if self.signing_key is None:
                raise NoSuitableSigningKeys

            jws = JWS(params, alg=self.jwt_algorithm)
            return jws.sign_compact(keys=[self.signing_key])
        except NoSuitableSigningKeys as signing_error:
            raise NoIssuerKey(
                "An issuer key wasn't loaded. "
                "Please run set_issuer() first.") from signing_error

    def _build_jwt_signature(self, method, resource, jti, subject,
                             content_hash=None):
        """
        Compiles a JWT signature for the given data
        :param method: HTTP Method
        :param content_hash: SHA 256 hash of the content body if it exists
        :param resource: The path (api endpoint)
        :param jti: JWT ID. This is a unique identifier for the request
        :param subject: The subject entity of the request
        :return:
        """
        current = int(time()) - self.server_time_difference
        expires = current + 5

        params = {
            "alg": self.jwt_algorithm,
            "nbf": current,
            "exp": expires,
            "iss": self.issuer,
            "sub": subject,
            "aud": self.audience,
            "iat": current,
            "jti": jti,
            "request": {
                "meth": method.upper(),
                "path": resource
            }
        }

        if content_hash is not None:
            params["request"]["hash"] = content_hash
            params["request"]["func"] = self.content_hash_algorithm

        return "IOV-JWT %s" % self._get_jwt_signature(params)

    def _get_jwt_payload(self, compact_jwt):
        try:
            return JWS().verify_compact(compact_jwt, keys=self.api_public_keys)
        except (AttributeError, BadSyntax) as jws_error:
            raise InvalidJWTResponse(
                "Received JWT is not valid: %s" % compact_jwt) from jws_error

    @staticmethod
    def _get_content_hash(body, hash_function):
        """
        Retrieves a hash using the stored content_hash_function
        :param body: string body
        :param hash_function: function identifier hash the body
        :return: hash based on the content_hash_function
        """
        if hash_function == "S256":
            hasher = sha256
        elif hash_function == "S384":
            hasher = sha384
        elif hash_function == "S512":
            hasher = sha512
        else:
            raise InvalidAlgorithm(
                "Invalid hash algorithm {}!".format(hash_function))

        return hasher(body.encode("latin1")).hexdigest()

    def _encrypt_request(self, data):
        """
        Encrypts the input data for the stored api_public_keys
        :param data: Information to be encrypted
        :return: JWE formatted string
        """
        jwe = JWE(json.dumps(data), alg=self.jwe_cek_encryption,
                  enc=self.jwe_claims_encryption)
        # Retrieve the active API encryption KID
        current_kid = self.update_and_return_active_encryption_kid()
        return jwe.encrypt(keys=[self._public_key_cache[current_kid]])

    def _process_jose_request(self, method, path, subject, data=None):
        """
        Performs a JOSE request
        :param method: Request method IE get, put, post, delete
        :param path: Path or endpoint that will be hit
        :param subject: Subject for which the request is issued for
        :param data: The data that will be submitted in the body of the request
        :return:
        """
        jti = self.__get_jti()
        body = None
        if data:
            body = self._encrypt_request(data)
            content_hash = self._get_content_hash(body,
                                                  self.content_hash_algorithm)
            signature = self._build_jwt_signature(method, path, jti, subject,
                                                  content_hash=content_hash)
            headers = {"content-type": "application/jwe",
                       "Authorization": signature}
        else:
            signature = self._build_jwt_signature(method, path, jti, subject)
            headers = {"content-type": "application/jwt",
                       "Authorization": signature}
        headers["User-Agent"] = f"PythonServiceSDK/{SDK_VERSION} " \
                                f"({platform.system()} {platform.release()})"
        response = getattr(self._http_client, method.lower())(path, data=body,
                                                              headers=headers)

        if response.status_code != 401:
            self.verify_jwt_response(response.headers, jti, response.data,
                                     subject)

        if response.data and not isinstance(response.data, dict):
            jwe = self.decrypt_response(response.data)
            try:
                result = json.loads(jwe)
            except (ValueError, TypeError):
                result = jwe
            response.data = result

        if isinstance(response, APIErrorResponse):
            raise LaunchKeyAPIException(response.data, response.status_code,
                                        response.reason)

        return response

    def decrypt_response(self, response):
        """
        Decrypts a response using the stored issuer private keys
        :param response: JWE encrypted string
        :return: Decrypted string
        """
        package = JWEnc().unpack(response)
        keys = []
        if 'kid' in package.headers:
            for key in self.issuer_private_keys:
                if key.kid == package.headers['kid']:
                    keys = [key]
            if len(keys) == 0:
                raise EntityKeyNotFound("The key id: %s could not be found in "
                                        "the entities available keys."
                                        % package.headers["kid"])
        else:
            keys = list(self.issuer_private_keys)
        return JWE().decrypt(response, keys=keys).decode('utf-8')

    def decrypt_rsa_response(self, response, key_id):
        """
        Decrypts a RSA response using the stored issuer private keys
        :param response: RSA encrypted string
        :param key_id: Key ID designating who the response was encrypted for.
        :return: Decrypted string
        """
        if key_id not in self.loaded_issuer_private_keys:
            raise UnexpectedKeyID("The response was for a key id "
                                  "%s which is not recognized" %
                                  key_id)
        binary_package = b64decode(response)
        return self.loaded_issuer_private_keys[key_id].decrypt(binary_package)

    def verify_jwt_response(self, headers, jti, content_body, subject,
                            status_code=None):
        """
        Verifies a response's JWT header to be valid
        :param headers: Full headers of the response
        :param jti: The input JTI value in the initial request that returned
        the response
        :param content_body: The body of the response
        :param subject: Expected subject of the JWT
        :param status_code: Status code of the response
        :return: The JWT claims
        """
        if status_code is None:
            warnings.warn(
                "Not passing a status_code value has been deprecated and "
                "will not be allowed in the next major version",
                DeprecationWarning)

        ci_headers = {k.lower(): v for k, v in headers.items()}
        auth = headers.get('X-IOV-JWT')

        # Ensure key exists in cache, fetch from API by ID otherwise
        kid = self.__get_kid_from_api_response_headers(headers)
        public_key = self._find_key_by_kid(kid)
        self._cache_public_key(kid, public_key)

        try:
            payload = self._get_jwt_payload(auth)
        except JWKESTException as reason:
            raise JWTValidationFailure("Unable to parse JWT",
                                       reason=reason) from reason

        self._verify_jwt_payload(payload, self.audience, self.issuer, subject)

        if payload.get("jti") != jti:
            raise JWTValidationFailure(
                "JTI does not match: expected {0} but got {1}".format(
                    jti, payload.get("jti")))
        if not payload.get('response'):
            raise JWTValidationFailure("Expected JWT to contain a response "
                                       "segment but there was none!")

        response = payload['response']
        if 'status' not in response:
            raise JWTValidationFailure("Expected JWT to contain status "
                                       "in the response segment but there "
                                       "was none!")
        if status_code and status_code != response['status']:
            raise JWTValidationFailure("Unexpected response status value")

        self._verify_jwt_response_headers(ci_headers, response)
        self._verify_jwt_content_hash(content_body, response)
        return payload

    @staticmethod
    def _verify_jwt_response_headers(ci_headers, response):
        if 'location' in response and 'location' not in ci_headers:
            raise JWTValidationFailure(
                "Expected headers to location but there was none!")
        if 'location' not in response and 'location' in ci_headers:
            raise JWTValidationFailure("Expected JWT to contain location "
                                       "in the response segment but there "
                                       "was none!")
        if 'location' in response and 'location' in ci_headers \
                and response['location'] != ci_headers['location']:
            raise JWTValidationFailure("Invalid location header!")
        if 'cache' in response and 'cache-control' not in ci_headers:
            raise JWTValidationFailure("Expected headers to contain "
                                       "cache-control but there was none!")
        if 'cache' not in response and 'cache-control' in ci_headers:
            raise JWTValidationFailure("Expected JWT to contain cache in the "
                                       "response segment but there was none!")
        if 'cache' in response and 'cache-control' in ci_headers and \
                response['cache'] != ci_headers['cache-control']:
            raise JWTValidationFailure("Invalid cache-control header!")

    def verify_jwt_request(self, compact_jwt, subject, method, path, body):
        """
        Verify a request's JWT
        :param compact_jwt:  The compact JWT to verify
        :param subject: The expected subject of the JWT
        :param method: The method of the request
        :param path: The path of the request
        :param body: The body of the request
        :return: The claims of the JWT
        """
        try:
            # Ensure key exists in cache, fetch from API by ID otherwise
            kid = JWT().unpack(compact_jwt).headers["kid"]
            public_key = self._find_key_by_kid(kid)
            self._cache_public_key(kid, public_key)
            payload = self._get_jwt_payload(compact_jwt)
        except JWKESTException as reason:
            raise JWTValidationFailure("Unable to parse JWT",
                                       reason=reason) from reason

        self._verify_jwt_payload(payload, self.audience, self.issuer, subject)

        if 'request' not in payload:
            raise JWTValidationFailure("Expected JWT to contain a request "
                                       "segment but there was none!")
        if 'meth' not in payload['request']:
            raise JWTValidationFailure("Expected method attribute but "
                                       "there was none!")
        if method is not None and payload['request']['meth'] != method:
            raise JWTValidationFailure(
                "Method does not match: expected {} but got {}".format(
                    payload['request']['meth'], method))
        if 'path' not in payload['request']:
            raise JWTValidationFailure(
                "Expected path attribute but there was none!")
        if path is not None and payload['request']['path'] != path:
            raise JWTValidationFailure(
                "Path does not match: expected {} but got {}".format(
                    payload['request']['path'], path))

        self._verify_jwt_content_hash(body, payload['request'])
        return payload

    @staticmethod
    def _verify_jwt_payload(payload, issuer, audience, subject):
        """
        Verifies a JWT's payload to be valid
        :param payload: JTW payload to verify
        :param issuer: Expected issuer of the JWT payload
        :param audience: Expected audience of the JWT payload
        :param subject: Expected subject of the JWT payload
        :return: The JWT claims
        """
        now = time()

        if payload.get('iss') != issuer:
            raise JWTValidationFailure(
                "Issuer does not match: expected %s but got %s" % (
                    issuer, payload.get('iss')))
        if payload.get('aud') != audience:
            raise JWTValidationFailure(
                "Audience does not match: expected %s but got %s" % (
                    audience, payload.get('aud')))
        if payload.get('nbf') > now + JOSE_JWT_LEEWAY:
            seconds = payload.get('nbf') - (now + JOSE_JWT_LEEWAY)
            raise JWTValidationFailure("NBF failed by %s seconds" % seconds)
        if payload.get('exp') < now - JOSE_JWT_LEEWAY:
            seconds = (now - JOSE_JWT_LEEWAY) - payload.get('exp')
            raise JWTValidationFailure("EXP failed by %s seconds" % seconds)
        if payload.get('sub') != subject:
            raise JWTValidationFailure(
                "Subject does not match: expected %s but got %s" % (
                    subject, payload.get('sub')))
        if payload.get('iat') > now + JOSE_JWT_LEEWAY:
            seconds = (payload.get('iat') - (now + JOSE_JWT_LEEWAY))
            raise JWTValidationFailure(
                "IAT failed as its %s seconds in the future" % seconds)

        return payload

    def _verify_jwt_content_hash(self, content, custom_segment):
        if not content and (
                custom_segment.get('hash') or custom_segment.get('func')):
            raise JWTValidationFailure(
                "JWT expected response body but was none!")
        if content and not custom_segment.get('hash'):
            raise JWTValidationFailure(
                "Expected JWT to contain a hash attribute but there was none!")
        if content and not custom_segment.get('func'):
            raise JWTValidationFailure(
                "Expected JWT to contain a func attribute but there was none!")
        if content:
            try:
                expected_hash = self._get_content_hash(content,
                                                       custom_segment.get(
                                                           'func'))
            except InvalidAlgorithm as reason:
                raise JWTValidationFailure("Invalid algorithm in JWT",
                                           reason=reason) from reason
            received_hash = custom_segment.get('hash')
            if received_hash != expected_hash:
                raise JWTValidationFailure(
                    "Content hash does not match: expected %s but got %s" %
                    (expected_hash, received_hash))

    def get(self, path, subject=None, **kwargs):
        """
        Performs an HTTP GET request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param subject: Subject for which the request is issued for
        :param kwargs: Any additional KWARGs will be parameters that are
        tacked onto the url
        :return:
        """
        if subject:
            return self._process_jose_request('get', path, subject)
        return self._http_client.get(path, data=kwargs)

    def post(self, path, subject=None, **kwargs):
        """
        Performs an HTTP POST request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param subject: Subject for which the request is issued for
        :param kwargs: Any additional KWARGs will be converted to data
        parameters
        :return:
        """
        return self._process_jose_request('post', path, subject, kwargs)

    def put(self, path, subject=None, **kwargs):
        """
        Performs an HTTP PUT request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param subject: Subject for which the request is issued for
        :param kwargs: Any additional KWARGs will be converted to data
        parameters
        :return:
        """
        return self._process_jose_request('put', path, subject, kwargs)

    def delete(self, path, subject=None, **kwargs):
        """
        Performs an HTTP DELETE request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param subject: Subject for which the request is issued for
        :param kwargs: Any additional KWARGs will be converted to data
        parameters
        :return:
        """
        return self._process_jose_request('delete', path, subject, kwargs)

    def patch(self, path, subject=None, **kwargs):
        """
        Performs an HTTP PATCH request against the LaunchKey API
        :param path: Path or endpoint that will be hit
        :param subject: Subject for which the request is issued for
        :param kwargs: Any additional KWARGs will be converted to data
        parameters
        :return:
        """
        return self._process_jose_request('patch', path, subject, kwargs)



"""Exception Classes"""

# pylint: disable=keyword-arg-before-vararg


import warnings

from formencode import Invalid

from .validation import AuthorizationInProgressValidator


class InsufficientRights(Exception):
    """The requesting client does not have sufficient rights for a service"""


class InvalidEntityID(Exception):
    """The given entity ID is not a valid UUID"""


class InvalidPrivateKey(Exception):
    """The given private key is not a valid PEM string"""


class InvalidIssuer(Exception):
    """The issuer is not valid"""


class InvalidIssuerFormat(Exception):
    """The issuer format is not a valid UUID"""


class InvalidIssuerVersion(Exception):
    """The issuer UUID is the wrong version type"""


class InvalidAlgorithm(Exception):
    """Input algorithm is not supported"""


class InvalidPolicyFormat(Exception):
    """
    Invalid policy format. A JSON object is expected which is in the proper
    format containing minimum_requirements and factors.
    """


class DuplicateGeoFenceName(Exception):
    """The input Geo-Fence name is already taken"""


class InvalidGeoFenceName(Exception):
    """The input Geo-Fence name was not found"""


class InvalidFenceType(Exception):
    """The GeoFence provided was not allowed"""


class DuplicateTimeFenceName(Exception):
    """The input Time-Fence name is already taken"""


class InvalidTimeFenceName(Exception):
    """The input Time-Fence name was not found"""


class InvalidTimeFenceStartTime(Exception):
    """The input start time should be a datetime.time object"""


class InvalidTimeFenceEndTime(Exception):
    """The input end time should be a datetime.time object"""


class MismatchedTimeFenceTimezones(Exception):
    """Start time and end time timezones must match"""


class UnknownPolicyException(Exception):
    """Policy Type was unknown to this version of the SDK"""


class InvalidPolicyAttributes(Exception):
    """Invalid Policy Attributes on Inside or Outside Policies"""


class LaunchKeyAPIException(Exception):
    """API Error (400+) was returned"""

    def __init__(self, message=None, status_code=None, reason=None,
                 error_data=None, *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.message = message
        self.status_code = status_code
        self.reason = reason
        self.data = error_data if error_data else {}


class InvalidParameters(LaunchKeyAPIException):
    """API Error ARG-001 - Parameter validation error"""


class PolicyFailure(LaunchKeyAPIException):
    """
    API Error A-AUT-P-405 - Auth creation failed due to user not passing policy
    """


class InvalidPolicyInput(LaunchKeyAPIException):
    """API Error APP-002 - The input auth policy is not valid"""


class Unauthorized(LaunchKeyAPIException):
    """
    Generic API 401 Error - Authentication was denied. Likely an invalid key.
    """


class Forbidden(LaunchKeyAPIException):
    """Generic API 403 Error"""


class EntityNotFound(LaunchKeyAPIException):
    """Generic API 404 Error - Entity was not found"""


class RequestTimedOut(LaunchKeyAPIException):
    """Generic API 408 Error - Request timed out"""


class Conflict(LaunchKeyAPIException):
    """Generic API 409 Error - Conflict"""


class RateLimited(LaunchKeyAPIException):
    """Generic API 429 Error - Rate Limited"""


class InvalidDirectoryIdentifier(LaunchKeyAPIException):
    """API Error DIR-001 - The input directory identifier was invalid"""


class UnexpectedAPIResponse(LaunchKeyAPIException):
    """The API returned a response that cannot be handled"""


class UnexpectedDeviceResponse(LaunchKeyAPIException):
    """The device returned a response that cannot be handled"""


class InvalidDeviceStatus(LaunchKeyAPIException):
    """The device status code is not recognized"""


class JWTValidationFailure(LaunchKeyAPIException):
    """JWT Response from the API was determined to be invalid"""


class UnexpectedKeyID(LaunchKeyAPIException):
    """The given key id was not unrecognized"""


class NoIssuerKey(LaunchKeyAPIException):
    """Issuer key was not loaded"""


class EntityKeyNotFound(LaunchKeyAPIException):
    """The key id with this current response could not be found within the
    entities available keys"""


class InvalidJWTResponse(LaunchKeyAPIException):
    """JWT Response is not in a valid format"""


class InvalidRoute(LaunchKeyAPIException):
    """The requested route does not exist in the requested path + method"""


class ServiceNameTaken(LaunchKeyAPIException):
    """
    The requested Service name is already in use. Service names are unique.
    """


class ServiceNotFound(LaunchKeyAPIException):
    """
    The requested Service does not exist. Likely invalid UUID or the service
    has been removed.
    """


class AuthorizationInProgress(LaunchKeyAPIException):
    """
    An auth already exists for the given user. Another one cannot be created
    until the current one is either responded to or expires.

    Contains attributes to describe the blocking auth request:
        :attr my_auth: Boolean stating whether the auth belongs to the
        requesting Service or another one. This is necessary when the Service
        may want to resume the current auth request.
        :attr expires: datetime object stating when the auth request will
        expire, and thus another auth can be created without user intervention.
        :attr auth_request: Auth request ID associated with the blocking auth.
        This ID can be used to resume polling for an auth request.
    """

    from_same_service = None
    expires = None
    authorization_request_id = None

    def __init__(self, message=None, *args, **kwargs):
        super().__init__(message, *args, **kwargs)

        try:
            self.data = AuthorizationInProgressValidator().to_python(self.data)
            self.from_same_service = self.data['from_same_service']
            self.expires = self.data['expires']
            self.authorization_request_id = self.data['auth_request']
        except Invalid as exception:
            warnings.warn("Failed to parse AuthorizationInProgress data: "
                          "exception: {0} data: {1}".format(exception,
                                                            self.data))


class AuthorizationRequestCanceled(LaunchKeyAPIException):
    """
    The authorization request has been canceled.
    """


class AuthorizationResponseExists(LaunchKeyAPIException):
    """
    The authentication request has already been responded to so it cannot be
    modified.
    """


class PublicKeyAlreadyInUse(LaunchKeyAPIException):
    """
    The public key you supplied already exists for the requested entity. It
    cannot be added again.
    """


class InvalidPublicKey(LaunchKeyAPIException):
    """The public key you supplied is not valid."""


class PublicKeyDoesNotExist(LaunchKeyAPIException):
    """The key_id you supplied could not be found."""


class LastRemainingKey(LaunchKeyAPIException):
    """The last remaining public key cannot be removed."""


class LastRemainingSDKKey(LaunchKeyAPIException):
    """The last remaining Authenticator SDK Key cannot be removed."""


class InvalidSDKKey(LaunchKeyAPIException):
    """
    The Authenticator SDK Key you supplied does not belong to the referenced
    Directory
    """


class DirectoryNameInUse(LaunchKeyAPIException):
    """The input Directory name is already in use."""


class UnexpectedWebhookRequest(LaunchKeyAPIException):
    """There was an error processing the webhook request."""


class UnableToDecryptWebhookRequest(LaunchKeyAPIException):
    """There was an error decrypting the body of the webhook request."""


class UnableToDecryptAuthorizationResponse(LaunchKeyAPIException):
    """
    There was an error decrypting the response package of the authorization
    response webhook request.
    """


class UnexpectedAuthorizationResponse(LaunchKeyAPIException):
    """
    There was an issue with the response package of the authorization response
    webhook request.
    """


class WebhookAuthorizationError(LaunchKeyAPIException):
    """
    There was an issue with validating the authorization provided by the
    webhook request.
    """


class XiovJWTValidationFailure(LaunchKeyAPIException):
    """x-iov-jwt validation failure"""


class XiovJWTDecryptionFailure(LaunchKeyAPIException):
    """x-iov-jwt decryption failure"""



"""Directory Client module"""

# pylint: disable=too-many-arguments

import warnings
from json import loads
from formencode import Invalid
from ..entities.validation import DirectoryGetDeviceResponseValidator, \
    DirectoryGetSessionsValidator, DirectoryUserDeviceLinkResponseValidator, \
    DirectoryDeviceLinkCompletionValidator, DirectoryUserTOTPValidator
from ..entities.directory import Session, DirectoryUserDeviceLinkData, Device,\
    DeviceLinkCompletionResponse, DirectoryUserTOTP
from .base import ServiceManagingBaseClient, api_call
from ..exceptions import UnableToDecryptWebhookRequest, \
    UnexpectedWebhookRequest, XiovJWTValidationFailure, \
    XiovJWTDecryptionFailure
from ..utils.shared import XiovJWTService


class DirectoryClient(ServiceManagingBaseClient):
    """
    Client for interacting with Directory endpoints
    """

    def __init__(self, subject_id, transport):
        super().__init__('dir', subject_id, transport,
                         "/directory/v3/services")
        self.x_iov_jwt_service = XiovJWTService(self._transport, self._subject)

    @api_call
    def link_device(self, user_id, ttl=None):
        """
        Begin the process of Linking a Subscriber Authenticator Device with an
        End User based on the Directory User ID. If no Directory User exists
        for the Directory User ID, the Directory User will be created.
        :param user_id: Unique value identifying the End User in your
        system. It is the permanent link for the End
        User between the your application(s) and the LaunchKey API. This
        will be used for authorization requests as well as managing the End
        User's Devices.
        :param ttl: Number of seconds the linking code returned in the response
        should be valid. If no value is provided, the system default will be
        used.
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.InvalidDirectoryIdentifier - Input
        identifier is invalid.
        :return: launchkey.entities.directory.DirectoryUserDeviceLinkData -
        Contains data needed to complete the linking process
        """
        kwargs = {"identifier": user_id}
        if ttl is not None:
            kwargs['ttl'] = ttl

        response = self._transport.post("/directory/v3/devices",
                                        self._subject,
                                        **kwargs)
        data = self._validate_response(
            response,
            DirectoryUserDeviceLinkResponseValidator)
        return DirectoryUserDeviceLinkData(data)

    @api_call
    def get_linked_devices(self, user_id):
        """
        Get a list of Subscriber Authenticator Devices for a Directory User.
        If not Directory User exists for the Directory User ID, an empty
        list will be returned.
        :param user_id: Unique value identifying the End User in your
        system. This value was used to create the Directory User and Link
        Device.
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :return: List - An list of launchkey.entities.directory.Device objects
        for the specified user identifier.
        """
        response = self._transport.post("/directory/v3/devices/list",
                                        self._subject, identifier=user_id)

        devices = [
            Device(
                self._validate_response(d, DirectoryGetDeviceResponseValidator)
            ) for d in response.data
        ]
        return devices

    @api_call
    def unlink_device(self, user_id, device_id):
        """
        Unlink a users device
        :param user_id: Unique value identifying the End User in your
        system. This value was used to create the Directory User and Link
        Device.
        :param device_id: The unique identifier of the Device you wish to
        Unlink. It would be obtained via Device.id returned by
        get_linked_devices().
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.EntityNotFound - The input device was not
        found. It may already be unlinked.
        """
        self._transport.delete("/directory/v3/devices", self._subject,
                               identifier=user_id, device_id=str(device_id))

    @api_call
    def end_all_service_sessions(self, user_id):
        """
        End Service User Sessions for all Services in which a Session was
        started for the Directory User
        :param user_id: Unique value identifying the End User in your system.
        This value was used to create the Directory User and Link Device.
        :raise: launchkey.exceptions.InvalidParameters - Input parameters were
        not correct
        :raise: launchkey.exceptions.EntityNotFound - The user was not found.
        """
        self._transport.delete("/directory/v3/sessions", self._subject,
                               identifier=user_id)

    @api_call
    def get_all_service_sessions(self, user_id):
        """
        Retrieves all Service Sessions that belong to a User
        :param user_id: Unique value identifying the End User in your system.
        This value was used to create the Directory User and Link Device.
        :raise: launchkey.exceptions.EntityNotFound - The input user identifier
        does not exist in your directory, or it does not have any devices
        linked to it
        :return: List - launchkey.entities.directory.Session
        """
        sessions = []
        response = self._transport.post("/directory/v3/sessions/list",
                                        self._subject, identifier=user_id)
        for session in response.data:
            validated_data = self._validate_response(
                session, DirectoryGetSessionsValidator)

            session = Session(validated_data)
            sessions.append(session)

        return sessions

    @api_call
    def generate_user_totp(self, user_id):
        """
        Generates, adds, and returns a TOTP secret for a given user identifier.

        Note that a user can only have a single TOTP configured. Submitting
        this request when there is an existing configuration will overwrite any
        previous settings.

        :param user_id: Unique value identifying the End User in your
        system. This value was used to create the Directory User and Link
        Device.
        :return: launchkey.entities.directory.DirectoryUserTOTP
        """
        response = self._transport.post("/directory/v3/totp",
                                        self._subject, identifier=user_id)
        validated_data = self._validate_response(
            response, DirectoryUserTOTPValidator)
        return DirectoryUserTOTP(validated_data)

    @api_call
    def remove_user_totp(self, user_id):
        """
        Removes a TOTP configuration from a given user.
        :param user_id: Unique value identifying the End User in your
        system. This value was used to create the Directory User and Link
        Device.
        :return:
        """
        self._transport.delete("/directory/v3/totp", self._subject,
                               identifier=user_id)

    def handle_webhook(self, body, headers, method, path):
        """
        Handle a webhook callback
        :param body: The raw body that was send in the POST content
        :param headers: A generic map of response headers. These will be used
        to access and validate authorization
        :param path:  The path of the request
        :param method: The HTTP method of the request
        :return: launchkey.entities.directory.
        DirectoryUserDeviceLinkCompletionWebhookPackage
        :raises launchkey.exceptions.UnexpectedWebhookRequest: when the
        request or its cannot be parsed or fails
        validation.
        :raises launchkey.exceptions.UnableToDecryptWebhookRequest: when the
        request is an authorization response webhook and the request body
        cannot be decrypted
        :raises launchkey.exceptions.UnexpectedKeyID: when the auth package in
        an authorization response webhook request body is decrypted using a
        public key whose private key is not known by the client. This can be
        a configuration issue.
        :raises launchkey.exceptions.WebhookAuthorizationError: when the
        "Authorization" header in the headers.
        """
        result = None
        try:
            decrypted_body = self.x_iov_jwt_service.decrypt_jwe(body, headers,
                                                                method, path)
            payload = loads(decrypted_body)
            device_link_data = DirectoryDeviceLinkCompletionValidator(
            ).to_python(payload)
            result = DeviceLinkCompletionResponse(
                device_link_data
            )
        except Invalid:
            warnings.warn("Invalid Directory Webhook received. There may be"
                          " an update available to add support.")
        except XiovJWTDecryptionFailure as reason:
            raise UnableToDecryptWebhookRequest(reason=reason) from reason
        except XiovJWTValidationFailure as reason:
            raise UnexpectedWebhookRequest(reason) from reason

        return result



""" Shared Utilities """

from functools import wraps
from uuid import UUID
import warnings

from jwkest import JWKESTException

from ..exceptions import InvalidIssuerFormat, InvalidIssuerVersion, \
    JWTValidationFailure, InvalidJWTResponse, WebhookAuthorizationError, \
    XiovJWTValidationFailure, XiovJWTDecryptionFailure


class XiovJWTService(object):
    """
    Handles the x-iov-jwt request spec validation and decryption
    """

    def __init__(self, transport, subject):
        self._transport = transport
        self._subject = subject

    def verify_jwt_request(self, body, headers, method, path):
        """
        Retrieves and validates an x-iov-jwt payload
        :param body: The raw body that was send in the POST content
        :param headers: A generic map of response headers. These will be used
        to access and validate authorization
        :param method: The HTTP method of the request
        :param path:  The path of the request
        :return: utf-8 encoded string of the body
        :raises launchkey.exceptions.XiovJWTValidationFailure: when the
        request or its cannot be parsed or fails
        validation.
        :raises launchkey.exceptions.WebhookAuthorizationError: when the
        "Authorization" header in the headers.
        """
        if not isinstance(body, str):
            body = body.decode("utf-8")

        compact_jwt = None
        for header_key, header_value in headers.items():
            if header_key.lower() == 'x-iov-jwt':
                compact_jwt = header_value

        if compact_jwt is None:
            raise WebhookAuthorizationError(
                "The X-IOV-JWT header was not found in the supplied headers "
                "from the request!")

        try:
            self._transport.verify_jwt_request(
                compact_jwt,
                self._subject,
                method,
                path,
                body)
        except (JWTValidationFailure, InvalidJWTResponse) as reason:
            raise XiovJWTValidationFailure(reason=reason) from reason
        return body

    def decrypt_jwe(self, body, headers, method, path):
        """
        Verifies and decrypts a jwt request
        :param body: The raw body that was send in the POST content
        :param headers: A generic map of response headers. These will be used
        to access and validate authorization
        :param method: The HTTP method of the request
        :param path:  The path of the request
        :raises launchkey.exceptions.UnexpectedKeyID: when the request body is
        decrypted using a public key whose private key is not known by the
        client. This can be a configuration issue.
        :raises launchkey.exceptions.XiovJWTDecryptionFailure: when the request
        body cannot be decrypted.
        :return: Decrypted string
        """
        body = self.verify_jwt_request(body, headers, method, path)
        try:
            return self._transport.decrypt_response(body)
        except JWKESTException as reason:
            raise XiovJWTDecryptionFailure(reason) from reason


class UUIDHelper(object):
    """
    Validate the provided uuid string and return a UUID if string is a valid
    UUID with the correct version or throw InvalidIssuerFormat or
    InvalidIssuerVersion when that criteria is not met.
    """

    def from_string(self, uuid_value, version=None):
        """
        Create a UUID from its string representation
        :param uuid_value: The string representation of a UUID
        :param version: The version of the UUID
        :return: UUID
        :raises launchkey.exceptions.InvalidIssuerFormat: when uuid_value
        is not a valid UUID format
        :raises launchkey.exceptions.InvalidIssuerVersion: when uuid_value
        is not the same version as version.
        """
        if not isinstance(uuid_value, UUID):
            try:
                uuid_value = UUID(uuid_value)
            except (ValueError, TypeError, AttributeError):
                raise InvalidIssuerFormat() from None

        self.validate_version(uuid_value, version)
        return uuid_value

    @staticmethod
    def validate_version(uuid_value, version):
        """
        Validate the the provided UUID is the provided version
        :param uuid_value: A UUID
        :param version: The expected version of the UUID
        :return: None
        :raises launchkey.exceptions.InvalidIssuerFormat: when uuid_value
        is not a UUID
        :raises launchkey.exceptions.InvalidIssuerVersion: when uuid_value
        is not the same version as version.
        """
        if not isinstance(uuid_value, UUID):
            raise InvalidIssuerFormat()

        if version is not None and int(version) != uuid_value.version:
            raise InvalidIssuerVersion()


def iso_format(datetime):
    """
    Generates an ISO formatted datetime based on what the LaunchKey API
    expects. This is a standard ISO datetime without microseconds.
    :param datetime: datetime.datetime object
    :return: ISO formatted string IE: 2017-10-03T22:50:15Z
    """
    return datetime.strftime("%Y-%m-%dT%H:%M:%SZ") \
        if datetime is not None else None


# pylint: disable=invalid-name
def deprecated(fn):
    """
    Decorator for issuing warnings
    :param fn: Function to be called
    :return Any: The expected return of the passed function
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        """ Decorator function """
        warnings.warn("The %s method has been deprecated and will be removed "
                      "in the next major release." % fn.__name__,
                      DeprecationWarning)

        return fn(*args, **kwargs)

    return wrapper



"""Base classes for factories"""

# pylint: disable=too-many-arguments,too-few-public-methods

import warnings
from ..transports import JOSETransport
from ..utils.shared import UUIDHelper


class BaseFactory(object):
    """
    LaunchKey Base client

    The primary purpose of the client is to validate entity input and
    translate needed values that will be utilized in
    the JOSE flow.
    """

    def __init__(self, issuer, issuer_id, private_key, url, testing,
                 transport):
        """
        :param issuer: Issuer type that will be translated directly to the
        JOSE transport layer as an issuer. IE: svc, dir, org
        :param issuer_id: UUID of the issuer
        :param private_key: PEM formatted private key string. This key will be
                            used for signing requests. It will also be used for
                            decrypting requests when a dual purpose key is
                            given. If a separate encryption key is desired it
                            can be added via the add_encryption_private_key
                            method.
        :param url: URL for the LaunchKey API
        :param testing: Boolean stating whether testing mode is being used.
        This will determine whether SSL validation
        occurs.
        """
        self._issuer_id = UUIDHelper().from_string(issuer_id)
        self._transport = transport if transport is not None \
            else JOSETransport()
        self._transport.set_url(url, testing)
        # Set the issue which will set the given key as the signature key
        self._transport.set_issuer(issuer, issuer_id, private_key)
        # Add the given key as an encryption key as well
        self._transport.add_encryption_private_key(private_key)

    def add_additional_private_key(self, private_key):
        """
        Adds an additional private key. This is to allow for key rotation.
        The default key to be used is the one set in
        the instantiation of the factory.

        This method is being deprecated in favor of add_encryption_private_key.

        :param private_key:
        :return:
        """
        warnings.warn(
            "This method will be removed in the future. Please use "
            "add_encryption_key instead.",
            DeprecationWarning)
        self.add_encryption_private_key(private_key)

    def add_encryption_private_key(self, private_key):
        """
        Adds an additional encryption private key. This is to allow for
        separating signature and encryption keys. Multiple encryption keys are
        supported to allow for key rotation.

        When defining a signature key it should be done in the instantiation of
        the factory.

        :param private_key:
        :return:
        """
        self._transport.add_encryption_private_key(private_key)



"""Service Entity and sub entities"""

# pylint: disable=invalid-name,too-few-public-methods,too-many-arguments
# pylint: disable=redefined-builtin,too-many-instance-attributes
# pylint: disable=too-many-locals

import datetime
import warnings
from enum import Enum
from json import loads, dumps

import pytz
from formencode import Invalid

from launchkey.entities.service.policy import AuthorizationResponsePolicy, \
    GeoCircleFence, TerritoryFence, Requirement
from launchkey.entities.validation import \
    AuthorizationResponsePackageValidator, \
    JWEAuthorizationResponsePackageValidator
from launchkey.exceptions import UnexpectedDeviceResponse, \
    InvalidGeoFenceName, InvalidTimeFenceEndTime, InvalidTimeFenceName, \
    InvalidTimeFenceStartTime, MismatchedTimeFenceTimezones, \
    DuplicateTimeFenceName, DuplicateGeoFenceName, InvalidPolicyFormat, \
    InvalidParameters


class AuthResponseReason(Enum):
    """Authentication Response Reason Enum"""
    APPROVED = "APPROVED"
    DISAPPROVED = "DISAPPROVED"
    FRAUDULENT = "FRAUDULENT"
    POLICY = "POLICY"
    PERMISSION = "PERMISSION"
    AUTHENTICATION = "AUTHENTICATION"
    CONFIGURATION = "CONFIGURATION"
    BUSY_LOCAL = "BUSY_LOCAL"
    SENSOR = "SENSOR"
    OTHER = "OTHER"


class AuthResponseType(Enum):
    """Authentication Response Type Enum"""
    AUTHORIZED = "AUTHORIZED"
    DENIED = "DENIED"
    FAILED = "FAILED"
    OTHER = "OTHER"


class AuthMethodType(Enum):
    """Authentication Request Method Type Enum"""
    PIN_CODE = "PIN_CODE"
    CIRCLE_CODE = "CIRCLE_CODE"
    GEOFENCING = "GEOFENCING"
    LOCATIONS = "LOCATIONS"
    WEARABLES = "WEARABLES"
    FINGERPRINT = "FINGERPRINT"
    FACE = "FACE"
    OTHER = "OTHER"


class GeoFence(object):
    """Geo-Fence entity"""
    def __init__(self, latitude, longitude, radius, name):
        warnings.warn(
            "GeoFence has been deprecated and will be removed in the next "
            "major version.", category=DeprecationWarning)
        self.latitude = float(latitude)
        self.longitude = float(longitude)
        self.radius = float(radius)
        self.name = str(name) if name else None

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, GeoFence):
            eq = self.name == other.name and \
                      self.latitude == other.latitude and \
                      self.longitude == other.longitude and \
                      self.radius == other.radius
        else:
            eq = False
        return eq

    def __repr__(self):
        return "GeoFence <" \
               "name=\"{name}\", " \
               "latitude={latitude}, " \
               "longitude={longitude}, " \
               "radius={radius}>". \
            format(
                name=self.name,
                latitude=self.latitude,
                longitude=self.longitude,
                radius=self.radius
            )


class TimeFence(object):
    """Time-Fence entity"""
    def __init__(self, name, start_time, end_time, monday=False, tuesday=False,
                 wednesday=False, thursday=False, friday=False,
                 saturday=False, sunday=False):

        warnings.warn(
            "TimeFence has been deprecated and will be removed in the next "
            "major version.", category=DeprecationWarning)

        if not isinstance(start_time, datetime.time):
            raise InvalidTimeFenceStartTime
        if not isinstance(end_time, datetime.time):
            raise InvalidTimeFenceEndTime
        if start_time.tzname() != end_time.tzname():
            raise MismatchedTimeFenceTimezones

        self.name = name
        self.start_time = start_time
        self.end_time = end_time
        self.monday = monday
        self.tuesday = tuesday
        self.wednesday = wednesday
        self.thursday = thursday
        self.friday = friday
        self.saturday = saturday
        self.sunday = sunday
        self.days = []
        if self.monday:
            self.days.append("Monday")
        if self.tuesday:
            self.days.append("Tuesday")
        if self.wednesday:
            self.days.append("Wednesday")
        if self.thursday:
            self.days.append("Thursday")
        if self.friday:
            self.days.append("Friday")
        if self.saturday:
            self.days.append("Saturday")
        if self.sunday:
            self.days.append("Sunday")
        self.timezone = start_time.tzname() if start_time.tzname() else "UTC"

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, TimeFence):
            eq = self.name == other.name and \
                      self.start_time == other.start_time and \
                      self.end_time == other.end_time and \
                      self.monday == other.monday and \
                      self.tuesday == other.tuesday and \
                      self.wednesday == other.wednesday and \
                      self.thursday == other.thursday and \
                      self.friday == other.friday and \
                      self.saturday == other.saturday and \
                      self.sunday == other.sunday and \
                      self.days == other.days
        else:
            eq = False
        return eq

    def __repr__(self):
        return "TimeFence <" \
               "name=\"{name}\", " \
               "start_time=\"{start_time}\", " \
               "end_time=\"{end_time}\", " \
               "monday={monday}, " \
               "tuesday={tuesday}, " \
               "wednesday={wednesday}, " \
               "thursday={thursday}, " \
               "friday={friday}, " \
               "saturday={saturday}, " \
               "sunday={sunday}>".\
            format(
                name=self.name,
                start_time=self.start_time,
                end_time=self.end_time,
                monday=self.monday,
                tuesday=self.tuesday,
                wednesday=self.wednesday,
                thursday=self.thursday,
                friday=self.friday,
                saturday=self.saturday,
                sunday=self.sunday
            )

    def __iter__(self):
        yield "days", self.days
        yield "start hour", self.start_time.hour
        yield "end hour", self.end_time.hour
        yield "name", self.name
        yield "start minute", self.start_time.minute
        yield "end minute", self.end_time.minute
        yield "timezone", self.timezone


class DenialReason(object):
    """
    Denial reason object to be passed in a list when creating an auth request

    :param denial_id: String. Universally unique identifier for this reason.
    This must be unique in an authorization request and should be unique
    globally for proper tracking of user selections.
    :param reason: String. Textual description of the reason that will be
    presented in a list for the user to select.
    :param fraud: Bool. Is the reason considered fraud.
    """

    def __init__(self, denial_id, reason, fraud):
        self.denial_id = denial_id
        self.reason = reason
        self.fraud = fraud


class AuthPolicy(object):
    """AuthPolicy object for performing dynamic authorizations"""

    def __init__(self, any=0, knowledge=False, inherence=False,
                 possession=False, jailbreak_protection=False):
        """
        Note that if any is used neither knowledge, inherence, nor possession
        can be used alongside it.

        If all values are left at 0, service defaults will be used.

        :param any: Int. Whether to require any x number of factors
        :param knowledge: Boolean. Whether to require knowledge factor
        :param inherence: Boolean. Whether to require inherence factor
        :param possession: Boolean. Whether to require possesion factor
        :param jailbreak_protection: Boolean. Whether to allow jailbroken /
               rooted devices to authenticate
        """

        warnings.warn("This object has been deprecated and will be removed in "
                      "the next major version. "
                      "Please use one of the new Policy objects in the "
                      "policy submodule",
                      category=DeprecationWarning)

        if knowledge not in (True, False, 0, 1) \
                or inherence not in (True, False, 0, 1) \
                or possession not in (True, False, 0, 1):
            raise InvalidParameters("Inherence, knowledge, or possesion "
                                    "]input must be a boolean.")
        if any != 0 and (knowledge or inherence or possession):
            raise InvalidParameters("Cannot use \"any\" with other specific "
                                    "factor requirements")

        self.geofences = []
        self.minimum_requirements = []
        self.minimum_amount = 0

        self._policy = {"factors": []}
        self.set_minimum_requirements(knowledge, inherence, possession, any)
        self.jailbreak_protection = jailbreak_protection
        self.require_jailbreak_protection(jailbreak_protection)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, AuthPolicy):
            eq = self.get_policy() == other.get_policy()
        else:
            eq = False
        return eq

    def add_geofence(self, latitude, longitude, radius, name=None):
        """
        Adds a Geo-Fence requirement
        :param latitude: Float. Geographical Latitude
        :param longitude: Float. Geographical Longitude
        :param radius: Float. Radius of the Geo-Fence in meters
        :param name: String. Optional identifier for the Geo-Fence.
        """
        try:
            geofence = GeoFence(latitude, longitude, radius, name)
            self.geofences.append(geofence)
            location = {
                "radius": geofence.radius,
                "latitude": geofence.latitude,
                "longitude": geofence.longitude
            }
            if geofence.name is not None:
                location['name'] = geofence.name
        except (TypeError, ValueError):
            raise InvalidParameters("Latitude, Longitude, and Radius "
                                    "must all be numbers.") from None
        for i, factor in enumerate(self._policy['factors']):
            if factor.get('factor') == "geofence":
                self._policy['factors'][i]['attributes']['locations'].append(
                    location)
                return
        self._policy['factors'].append(
            {
                "factor": "geofence",
                "requirement": "forced requirement",
                "quickfail": False,
                "priority": 1,
                "attributes": {"locations": [location]}
            }
        )

    def remove_geofence(self, name):
        """
        Removes a Geo-Fence based on an input name
        :param name: String name of the Geo-Fence
        :return:
        """
        def _remove_from_factors(name_):
            for key, factor in enumerate(self._policy['factors']):
                if 'factor' in factor and factor['factor'] == 'geofence':
                    for loc_key, location in enumerate(
                            factor['attributes']['locations']):
                        if 'name' in location \
                                and location['name'].lower() == name_.lower():
                            del self._policy['factors'][key]['attributes'][
                                'locations'][loc_key]
                            return True
            return False

        def _remove_from_fences(name):
            for geo_key, geo in enumerate(self.geofences):
                if geo.name.lower() == name.lower():
                    del self.geofences[geo_key]
                    return True
            return False

        if _remove_from_factors(name) and _remove_from_fences(name):
            return

        raise InvalidGeoFenceName

    def require_jailbreak_protection(self, status):
        """
        Enables or disables jailbreak and root protection (device integrity)
        for Auths
        :param status: Bool as to whether device integrity should be required
        :return:
        """
        enabled = 1 if status else 0
        self.jailbreak_protection = enabled == 1

        for key, factor in enumerate(self._policy['factors']):
            if 'factor' in factor and factor['factor'] == 'device integrity':
                self._policy['factors'][key]['attributes'] = \
                    {"factor enabled": 1 if status else 0}
                return

        self._policy['factors'].append({
            "factor": "device integrity",
            "requirement": "forced requirement",
            "quickfail": False,
            "priority": 1,
            "attributes": {"factor enabled": enabled}
        })

    def set_minimum_requirements(self, knowledge=False, inherence=False,
                                 possession=False, minimum_amount=0):
        """
        Sets minimum requirements that must be used for each Auth Request
        :param knowledge: Bool. Whether a Knowledge factor should be required.
        :param inherence: Bool. Whether an Inherence factor should be required.
        :param possession: Bool. Whether a Possession factor should be
        required.
        :param minimum_amount: Integer. Whether to require a minimum amount
        any nonspecific requirements.
        :return:
        """

        minimum_requirements = []
        if knowledge:
            minimum_requirements.append("knowledge")
        if inherence:
            minimum_requirements.append("inherence")
        if possession:
            minimum_requirements.append("possession")

        self.minimum_requirements = minimum_requirements
        self.minimum_amount = minimum_amount
        if not minimum_requirements and not minimum_amount:
            self._policy['minimum_requirements'] = []
        else:
            self._policy['minimum_requirements'] = [{
                "requirement": "authenticated",
                "any": self.minimum_amount,
                "knowledge": 1 if 'knowledge' in minimum_requirements else 0,
                "inherence": 1 if 'inherence' in minimum_requirements else 0,
                "possession": 1 if 'possession' in minimum_requirements else 0,
            }]

    def get_policy(self):
        """
        Retrieves json representation of the policy
        :return: Valid policy dict
        """
        warnings.warn("This method has been deprecated and will be removed in "
                      "the next major version. Please use "
                      "to_dict",
                      category=DeprecationWarning)
        return self.to_dict()

    def to_dict(self):
        """
        returns the JSON representation of the policy object
        """
        try:
            dumps(self._policy)
        except TypeError:
            raise InvalidParameters("Policy input was not JSON serializable. "
                                    "Please verify it is correct.") from None
        return self._policy

    def set_policy(self, policy):
        """
        Updates a policy based on a given policy
        :param policy: Dict os JSON representation of a LaunchKey policy
        :return: None
        """
        if isinstance(policy, dict):
            self._policy = policy
        else:
            try:
                self._policy = loads(policy)
            except (ValueError, TypeError):
                raise InvalidPolicyFormat() from None

        if 'minimum_requirements' not in self._policy \
                or 'factors' not in self._policy:
            raise InvalidPolicyFormat()

        self._parse_minimum_requirements(self._policy['minimum_requirements'])
        self._parse_factors(self._policy['factors'])

    def _parse_minimum_requirements(self, minimum_requirements):
        # Although the API returns a list, we only support one
        # returned value as of now
        if minimum_requirements:
            requirement = minimum_requirements[0]
            if 'knowledge' in requirement and requirement['knowledge']:
                self.minimum_requirements.append('knowledge')
            if 'inherence' in requirement and requirement['inherence']:
                self.minimum_requirements.append('inherence')
            if 'possession' in requirement and requirement['possession']:
                self.minimum_requirements.append('possession')
            if 'all' in requirement and requirement['all']:
                self.minimum_amount = requirement['all']
            if 'any' in requirement and requirement['any']:
                self.minimum_amount = requirement['any']

    def _parse_factors(self, factors):
        for factor in factors:
            if 'factor' in factor and 'attributes' in factor \
                    and factor['attributes']:
                if factor['factor'] == 'geofence' \
                        and 'locations' in factor['attributes']:
                    for fence in factor['attributes']['locations']:
                        self.geofences.append(
                            GeoFence(fence['latitude'], fence['longitude'],
                                     fence['radius'],
                                     fence.get('name', None)))
                elif factor['factor'] == 'device integrity' \
                        and 'factor enabled' in factor['attributes']:
                    if factor['attributes']['factor enabled']:
                        self.jailbreak_protection = True

    def __repr__(self):
        return "AuthPolicy <" \
               "minimum_requirements={minimum_requirements}, " \
               "minimum_amount={minimum_amount}, " \
               "geofences={geofences}>".\
            format(
                minimum_requirements=self.minimum_requirements,
                minimum_amount=self.minimum_amount,
                geofences=self.geofences
            )


class AuthorizationRequest(object):
    """
    Authorization Request object containing decrypted auth response
    and other related information
    """

    def __init__(self, auth_request, push_package, device_ids=None):
        self.auth_request = auth_request
        self.push_package = push_package
        self.device_ids = device_ids

    def __repr__(self):
        return "AuthorizationRequest <" \
               "auth_request=\"{auth_request}\", " \
               "push_package=\"{push_package}\", " \
               "device_ids=\"{device_ids}\">".\
            format(
                auth_request=self.auth_request,
                push_package=self.push_package,
                device_ids=self.device_ids
            )


class AuthMethod(object):
    """
    Auth method that describes the state of an object the policy utilized by
    the device while processing the identified authorization request.
    """

    def __init__(self, method, set, active, allowed, supported, user_required,
                 passed, error):
        self.method = method
        self.set = set
        self.active = active
        self.allowed = allowed
        self.supported = supported
        self.user_required = user_required
        self.passed = passed
        self.error = error

    def __repr__(self):
        return "AuthMethod <" \
               "method={method}, " \
               "set={set}, " \
               "active={active}, " \
               "allowed={allowed}, " \
               "supported={supported}, " \
               "user_required={user_required}, " \
               "passed={passed}, " \
               "error={error}>".\
            format(
                method=self.method.value if self.method else None,
                set=self.set,
                active=self.active,
                allowed=self.allowed,
                supported=self.supported,
                user_required=self.user_required,
                passed=self.passed,
                error=self.error
            )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, AuthMethod):
            eq = self.method == other.method and \
                 self.set == other.set and \
                 self.active == other.active and \
                 self.allowed == other.allowed and \
                 self.supported == other.supported and \
                 self.user_required == other.user_required and \
                 self.passed == other.passed and \
                 self.error == other.error
        else:
            eq = False
        return eq


class AdvancedAuthorizationResponse(object):
    """
    Authorization Response object containing decrypted auth response and
    other related information
    """

    @staticmethod
    def _get_authorized_bool_from_auth_response_context(response_type,
                                                        response_reason):
        return response_type == AuthResponseType.AUTHORIZED and \
               response_reason == AuthResponseReason.APPROVED

    @staticmethod
    def _retrieve_enum_from_value(enum_class, value):
        try:
            return enum_class(value)
        except ValueError:
            return enum_class.OTHER

    @staticmethod
    def __generate_fence_from_policy_dict(fence):
        if not fence.get("type"):
            # Denotes a legacy geofence that can become a GeoCircleFence
            return GeoCircleFence(
                fence["latitude"],
                fence["longitude"],
                fence["radius"],
                name=fence.get("name")
            )

        if fence["type"] == "TERRITORY":
            return TerritoryFence(
                fence["country"],
                administrative_area=fence.get("administrative_area"),
                postal_code=fence.get("postal_code"),
                name=fence.get("name")
            )

        if fence["type"] == "GEO_CIRCLE":
            return GeoCircleFence(
                fence["latitude"],
                fence["longitude"],
                fence["radius"],
                name=fence.get("name")
            )

        raise ValueError("Invalid fence type received in Auth Response")

    def _parse_and_set_auth_policy(self, auth_policy):
        kwargs = {}
        requirement = auth_policy["requirement"]

        try:
            if requirement:
                kwargs["requirement"] = Requirement(requirement.upper())

        except ValueError:
            kwargs["requirement"] = Requirement.OTHER

        if requirement == "amount":
            kwargs["amount"] = auth_policy["amount"]

        if requirement == "types":
            policy_types = map(lambda t: t.lower(), auth_policy["types"])
            kwargs["inherence_required"] = "inherence" in policy_types
            kwargs["knowledge_required"] = "knowledge" in policy_types
            kwargs["possession_required"] = "possession" in policy_types

        if auth_policy.get('geofences'):
            kwargs["fences"] = list(map(
                self.__generate_fence_from_policy_dict,
                auth_policy["geofences"]))

        self.policy = AuthorizationResponsePolicy(**kwargs)

    def _parse_device_response_from_jwe(self, jwe_payload, transport):
        """
        Parses a Device auth response using a JWE input.
        :param jwe_payload: JWE payload containing the device response.
        :param transport: Transport that contains the decrypt_response
                          method used to decrypt the Device response.
        :return:
        """
        try:
            data = transport.decrypt_response(jwe_payload)
            unmarshalled_package = loads(data)
            decrypted_jwe = JWEAuthorizationResponsePackageValidator().\
                to_python(unmarshalled_package)
        except (Invalid, TypeError, ValueError) as e:
            raise UnexpectedDeviceResponse("The device response was invalid. "
                                           "Please verify the same key that "
                                           "initiated the auth request is "
                                           "being used to decrypt the current "
                                           "message.", reason=e) from e

        self.type = self._retrieve_enum_from_value(AuthResponseType,
                                                   decrypted_jwe.get(
                                                       "type"))
        self.reason = self._retrieve_enum_from_value(AuthResponseReason,
                                                     decrypted_jwe.get(
                                                         "reason"))
        self.denial_reason = decrypted_jwe.get("denial_reason")
        self.fraud = self.reason is AuthResponseReason.FRAUDULENT
        self.authorization_request_id = decrypted_jwe.get("auth_request")
        self.authorized = \
            self._get_authorized_bool_from_auth_response_context(
                self.type,
                self.reason
            )
        self.device_id = decrypted_jwe.get("device_id")
        self.service_pins = decrypted_jwe.get("service_pins")

        auth_methods = decrypted_jwe.get("auth_methods")
        if auth_methods:
            self.auth_methods = [
                AuthMethod(
                    self._retrieve_enum_from_value(
                        AuthMethodType, method['method'].upper()),
                    method['set'],
                    method['active'],
                    method['allowed'],
                    method['supported'],
                    method['user_required'],
                    method['passed'],
                    method['error']
                )
                for method in auth_methods
            ]

        auth_policy = decrypted_jwe.get("auth_policy")
        if auth_policy:
            self._parse_and_set_auth_policy(auth_policy)

    def _parse_device_response_from_auth_package(self, auth_package, key_id,
                                                 transport):
        """
        Parses a Device auth response using a legacy auth package.
        :param auth_package: RSA Encrypted Device response.
        :param key_id: Key ID designating the key that the response
                       was encrypted to.
        :param transport: Transport that contains the _decrypt_auth_package
                          method used to decrypt the Device response.
        :return:
        """
        try:
            data = transport.decrypt_rsa_response(
                auth_package, key_id)
            unmarshalled_package = loads(data)
            # pylint: disable=no-value-for-parameter
            # noinspection PyArgumentList
            # The statement below should not work but does and will not work
            # when implemented properly. Disabling checks.
            decrypted_package = AuthorizationResponsePackageValidator.\
                to_python(unmarshalled_package)
            # pylint: enable=no-value-for-parameter
        except (Invalid, TypeError, ValueError) as e:
            raise UnexpectedDeviceResponse("The device response was invalid. "
                                           "Please verify the same key that "
                                           "initiated the auth request is "
                                           "being used to decrypt the current "
                                           "message.", reason=e) from e
        self.authorization_request_id = \
            decrypted_package.get("auth_request")
        self.authorized = decrypted_package.get("response")
        self.device_id = decrypted_package.get("device_id")
        self.service_pins = decrypted_package.get("service_pins")

    def _parse_device_response(self, data, transport):
        """
        Determines if a device has responded using JWE or legacy package format
        and triggers the appropriate parsing steps.
        :param data: Dictionary containing a LaunchKey Device's auth response
        :param transport: Transport that contains the needed steps to decrypt
                          A given Device response.
        :return: None
        """
        if data.get("auth_jwe"):
            self._parse_device_response_from_jwe(
                data.get("auth_jwe"), transport)
        else:
            self._parse_device_response_from_auth_package(
                data['auth'], data['public_key_id'], transport)

        self.service_user_hash = data.get("service_user_hash")
        self.organization_user_hash = data.get("org_user_hash")
        self.user_push_id = data.get("user_push_id")

    def __init__(self, data, transport):
        self.data = data
        self.transport = transport
        self.device_id = None
        self.type = None
        self.authorization_request_id = None
        self.service_pins = None
        self.authorized = None
        self.reason = None
        self.denial_reason = None
        self.fraud = None
        self.policy = None
        self.auth_methods = None
        self._parse_device_response(data, transport)


class AuthorizationResponse(AdvancedAuthorizationResponse):
    """
    Authorization Response object containing decrypted auth response and
    other related information
    """
    def __init__(self, data, transport):
        self.auth_policy = None
        super().__init__(data, transport)
        del self.policy

    def _parse_and_set_auth_policy(self, auth_policy):
        kwargs = {}
        if auth_policy['requirement'] == "amount":
            kwargs['any'] = auth_policy['amount']
        elif auth_policy['requirement'] == "types":
            for item in auth_policy['types']:
                type = item.lower()
                if type in ['knowledge', 'inherence', 'possession']:
                    kwargs[type] = True
                else:
                    warnings.warn(
                        "Invalid policy type given: %s. "
                        "It will be ignored, but this could "
                        "signify the need for an update." % type)
        if auth_policy.get('geofences') or kwargs:
            self.auth_policy = AuthPolicy(**kwargs)
            for fence in auth_policy.get('geofences', []):
                if not fence.get("type") or fence.get("type") == "GEO_CIRCLE":
                    self.auth_policy.add_geofence(
                        fence['latitude'],
                        fence['longitude'],
                        fence['radius'],
                        name=fence.get('name', None)
                    )
                else:
                    warnings.warn(
                        "A Fence besides GeoCircleFence was present"
                        " while using legacy functionality. This fence "
                        "has been skipped from being processed."
                    )


class SessionEndRequest(object):
    """
    Session end request containing the logout_requested unix timestamp
    and the service_user_hash
    """

    def __init__(self, service_user_hash, api_time):
        self.logout_requested = api_time
        self.service_user_hash = service_user_hash


class ServiceSecurityPolicy(AuthPolicy):
    """
    Security Policy object containing specifics on policy that will be
    used in Auth Requests
    """

    def __init__(self, any=0, knowledge=False, inherence=False,
                 possession=False, jailbreak_protection=False):
        self.timefences = []

        super().__init__(any, knowledge, inherence, possession,
                         jailbreak_protection)

    def add_geofence(self, latitude, longitude, radius, name=None):
        """
        Adds a Geo-Fence requirement
        :param latitude: Float. Geographical Latitude
        :param longitude: Float. Geographical Longitude
        :param radius: Float. Radius of the Geo-Fence in meters
        :param name: String. Name identifier for the Geo-Fence. This
        should be unique.
        """
        if name is None:
            raise ValueError("name expected not to be None!")
        for fence in self.geofences:
            if fence.name.lower() == name.lower():
                # If the name exists, we should raise an
                # error since they should to be unique
                raise DuplicateGeoFenceName
        return super().add_geofence(latitude, longitude, radius, name)

    def add_timefence(self, name, start_time, end_time, monday=False,
                      tuesday=False, wednesday=False, thursday=False,
                      friday=False, saturday=False, sunday=False):
        """
        Adds a Time-Fence requirement
        :param name: String. Name identifier for the Time-Fence. This should
        be unique.
        :param start_time: datetime.time object. A tzinfo value must be used if
        you do not want UTC. It must match the end_time tzinfo.
        :param end_time:  datetime.time object. A tzinfo value must be used if
        you do not want UTC. It must match the start_time tzinfo.
        :param monday: Bool. Whether the Time-Fence should be valid on Mondays
        :param tuesday: Bool. Whether the Time-Fence should be valid on
        Tuesdays
        :param wednesday: Bool. Whether the Time-Fence should be valid on
        Wednesdays
        :param thursday: Bool. Whether the Time-Fence should be valid on
        Thursdays
        :param friday: Bool. Whether the Time-Fence should be valid on Fridays
        :param saturday: Bool. Whether the Time-Fence should be valid on
        Saturdays
        :param sunday: Bool. Whether the Time-Fence should be valid on Sundays
        :return:
        """
        fence = TimeFence(name, start_time, end_time, monday=monday,
                          tuesday=tuesday, wednesday=wednesday,
                          thursday=thursday, friday=friday, saturday=saturday,
                          sunday=sunday)
        self.timefences.append(fence)
        new_time = {
            "days": fence.days,
            "start hour": fence.start_time.hour,
            "end hour": fence.end_time.hour,
            "name": fence.name,
            "start minute": fence.start_time.minute,
            "end minute": fence.end_time.minute,
            "timezone": fence.timezone
        }

        for key, factor in enumerate(self._policy['factors']):
            if 'factor' in factor and factor['factor'] == 'timefence':
                for fence in factor['attributes']['time fences']:
                    if fence['name'].lower() == name.lower():
                        # If the name exists, we should raise an error
                        # since they should to be unique
                        raise DuplicateTimeFenceName
                self._policy['factors'][key]['attributes']['time fences'] \
                    .append(new_time)
                return

        # If the timefence factor does not exist at all, we need to create it
        self._policy['factors'].append({
            "factor": "timefence",
            "requirement": "forced requirement",
            "quickfail": False,
            "priority": 1,
            "attributes": {
                "time fences": [
                    new_time
                ]
            }
        })

    def remove_timefence(self, name):
        """
        Removes a Time-Fence based on an input name
        :param name: String name of the Time-Fence
        :return:
        """
        def _remove_from_factors(name_):
            for key, factor in enumerate(self._policy['factors']):
                if 'factor' in factor and factor['factor'] == 'timefence':
                    for loc_key, location in \
                            enumerate(factor['attributes']['time fences']):
                        if location['name'].lower() == name_.lower():
                            del self._policy['factors'][key]['attributes'][
                                'time fences'][loc_key]
                            return True
            return False

        def _remove_from_fences(name_):
            for time_key, time in enumerate(self.timefences):
                if time.name.lower() == name_.lower():
                    del self.timefences[time_key]
                    return True
            return False

        if _remove_from_factors(name) and _remove_from_fences(name):
            return

        raise InvalidTimeFenceName

    def _parse_factors(self, factors):
        for factor in factors:
            if 'factor' in factor and 'attributes' in factor \
                    and factor['attributes']:
                if factor['factor'] == 'timefence' \
                        and 'time fences' in factor['attributes']:
                    for fence in factor['attributes']['time fences']:
                        # Dict comp to convert the days list into kwargs
                        kwargs = {day.lower(): True for day in fence['days']}
                        self.timefences.append(
                            TimeFence(
                                fence['name'],
                                datetime.time(
                                    hour=fence['start hour'],
                                    minute=fence['start minute'],
                                    tzinfo=pytz.timezone(fence['timezone'])),
                                datetime.time(
                                    hour=fence['end hour'],
                                    minute=fence['end minute'],
                                    tzinfo=pytz.timezone(fence['timezone'])),
                                **kwargs
                            )
                        )

        return super()._parse_factors(factors)


class Service(object):
    """
    LaunchKey Service. Can be either an Organization or Directory Service.
    """

    def __init__(self, service_data):
        self.id = service_data['id']
        self.icon = service_data['icon']
        self.name = service_data['name']
        self.description = service_data['description']
        self.active = service_data['active']
        self.callback_url = service_data['callback_url']



"""Service Factory"""

# pylint: disable=too-many-arguments

from .. import LAUNCHKEY_PRODUCTION
from ..clients import ServiceClient
from .base import BaseFactory


class ServiceFactory(BaseFactory):
    """
    Factory for creating clients when representing a LaunchKey Service Profile
    """

    def __init__(self, service_id, private_key, url=LAUNCHKEY_PRODUCTION,
                 testing=False, transport=None):
        """
        :param service_id: UUID for the requesting service
        :param private_key: PEM formatted private key string
        :param url: URL for the LaunchKey API
        :param testing: Boolean stating whether testing mode is being used.
        This will determine whether SSL validation occurs.
        :param: transport: Instantiated transport object. The default and
        currently only supported transport is
        launchkey.transports.JOSETransport. If you wish to set encryption or
        hashing algorithms, this is where you would do it. IE:
        JOSETransport(jwt_algorithm="RS512", jwe_cek_encryption="RSA-OAEP",
        jwe_claims_encryption="A256CBC-HS512", content_hash_algorithm="S256")
        """
        super().__init__('svc', service_id, private_key,
                         url, testing, transport)

    def make_service_client(self):
        """
        Retrieves a client to make service calls.
        :return: launchkey.clients.ServiceClient
        """
        return ServiceClient(self._issuer_id, self._transport)
